from typing import Sequence, Optional, Mapping

import pytest
from debian.deb822 import Deb822
from debian.debian_support import DpkgArchTable

from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.deb_packaging_support import (
    install_upstream_changelog,
    compute_installed_size,
    auto_compute_multi_arch,
    generate_md5sums_file,
)
from debputy.filesystem_scan import build_virtual_fs
from debputy.packages import BinaryPackage
from debputy.plugin.api import virtual_path_def
from debputy.plugin.api.spec import PathDef


@pytest.mark.parametrize(
    "upstream_changelog_name,other_files",
    [
        (
            "changelog.txt",
            [
                "changelog.md",
                "CHANGELOG.rst",
                "random-file",
            ],
        ),
        (
            "CHANGELOG.rst",
            [
                "doc/CHANGELOG.txt",
                "docs/CHANGELOG.md",
            ],
        ),
        (
            "docs/CHANGELOG.rst",
            [
                "docs/history.md",
            ],
        ),
        (
            "changelog",
            [],
        ),
    ],
)
def test_upstream_changelog_from_source(
    package_single_foo_arch_all_cxt_amd64,
    upstream_changelog_name,
    other_files,
) -> None:
    upstream_changelog_content = "Some upstream changelog"
    dctrl = package_single_foo_arch_all_cxt_amd64["foo"]
    data_fs_root = build_virtual_fs([], read_write_fs=True)
    upstream_fs_contents = [
        virtual_path_def("CHANGELOG", materialized_content="Some upstream changelog")
    ]
    upstream_fs_contents.extend(
        virtual_path_def(x, materialized_content="Wrong file!") for x in other_files
    )
    source_fs_root = build_virtual_fs(upstream_fs_contents)

    install_upstream_changelog(dctrl, data_fs_root, source_fs_root)

    upstream_changelog = data_fs_root.lookup(f"usr/share/doc/{dctrl.name}/changelog")
    assert upstream_changelog is not None
    assert upstream_changelog.is_file
    with upstream_changelog.open() as fd:
        content = fd.read()
    assert upstream_changelog_content == content


@pytest.mark.parametrize(
    "upstream_changelog_basename,other_data_files,other_source_files",
    [
        (
            "CHANGELOG",
            [
                "history.txt",
                "changes.md",
            ],
            [
                "changelog",
                "doc/CHANGELOG.txt",
                "docs/CHANGELOG.md",
            ],
        ),
        (
            "changelog",
            [
                "history.txt",
                "changes.md",
            ],
            [
                "changelog",
                "doc/CHANGELOG.txt",
                "docs/CHANGELOG.md",
            ],
        ),
        (
            "changes.md",
            [
                "changelog.rst",
            ],
            ["changelog"],
        ),
    ],
)
def test_upstream_changelog_from_data_fs(
    package_single_foo_arch_all_cxt_amd64,
    upstream_changelog_basename,
    other_data_files,
    other_source_files,
) -> None:
    upstream_changelog_content = "Some upstream changelog"
    dctrl = package_single_foo_arch_all_cxt_amd64["foo"]
    doc_dir = f"./usr/share/doc/{dctrl.name}"
    data_fs_contents = [
        virtual_path_def(
            f"{doc_dir}/{upstream_changelog_basename}",
            materialized_content="Some upstream changelog",
        )
    ]
    data_fs_contents.extend(
        virtual_path_def(
            f"{doc_dir}/{x}",
            materialized_content="Wrong file!",
        )
        for x in other_data_files
    )
    data_fs_root = build_virtual_fs(data_fs_contents, read_write_fs=True)
    source_fs_root = build_virtual_fs(
        [
            virtual_path_def(
                x,
                materialized_content="Wrong file!",
            )
            for x in other_source_files
        ]
    )

    install_upstream_changelog(dctrl, data_fs_root, source_fs_root)

    upstream_changelog = data_fs_root.lookup(f"usr/share/doc/{dctrl.name}/changelog")
    assert upstream_changelog is not None
    assert upstream_changelog.is_file
    with upstream_changelog.open() as fd:
        content = fd.read()
    assert upstream_changelog_content == content


def test_upstream_changelog_pre_installed_compressed(
    package_single_foo_arch_all_cxt_amd64,
) -> None:
    dctrl = package_single_foo_arch_all_cxt_amd64["foo"]
    changelog = f"./usr/share/doc/{dctrl.name}/changelog.gz"
    data_fs_root = build_virtual_fs(
        [virtual_path_def(changelog, fs_path="/nowhere/should/not/be/resolved")],
        read_write_fs=True,
    )
    source_fs_root = build_virtual_fs(
        [virtual_path_def("changelog", materialized_content="Wrong file!")]
    )

    install_upstream_changelog(dctrl, data_fs_root, source_fs_root)

    upstream_ch_compressed = data_fs_root.lookup(
        f"usr/share/doc/{dctrl.name}/changelog.gz"
    )
    assert upstream_ch_compressed is not None
    assert upstream_ch_compressed.is_file
    upstream_ch_uncompressed = data_fs_root.lookup(
        f"usr/share/doc/{dctrl.name}/changelog"
    )
    assert upstream_ch_uncompressed is None


def test_upstream_changelog_no_matches(
    package_single_foo_arch_all_cxt_amd64,
) -> None:
    dctrl = package_single_foo_arch_all_cxt_amd64["foo"]
    doc_dir = f"./usr/share/doc/{dctrl.name}"
    data_fs_root = build_virtual_fs(
        [
            virtual_path_def(
                f"{doc_dir}/random-file", materialized_content="Wrong file!"
            ),
            virtual_path_def(
                f"{doc_dir}/changelog.Debian", materialized_content="Wrong file!"
            ),
        ],
        read_write_fs=True,
    )
    source_fs_root = build_virtual_fs(
        [virtual_path_def("some-random-file", materialized_content="Wrong file!")]
    )

    install_upstream_changelog(dctrl, data_fs_root, source_fs_root)

    upstream_ch_compressed = data_fs_root.lookup(
        f"usr/share/doc/{dctrl.name}/changelog.gz"
    )
    assert upstream_ch_compressed is None
    upstream_ch_uncompressed = data_fs_root.lookup(
        f"usr/share/doc/{dctrl.name}/changelog"
    )
    assert upstream_ch_uncompressed is None


def test_upstream_changelog_salsa_issue_49(
    package_single_foo_arch_all_cxt_amd64,
) -> None:
    # https://salsa.debian.org/debian/debputy/-/issues/49
    dctrl = package_single_foo_arch_all_cxt_amd64["foo"]
    doc_dir_path = f"./usr/share/doc/{dctrl.name}"
    data_fs_root = build_virtual_fs(
        [virtual_path_def(doc_dir_path, link_target="foo-data")], read_write_fs=True
    )
    source_fs_root = build_virtual_fs(
        [virtual_path_def("changelog", materialized_content="Wrong file!")]
    )

    install_upstream_changelog(dctrl, data_fs_root, source_fs_root)

    doc_dir = data_fs_root.lookup(f"usr/share/doc/{dctrl.name}")
    assert doc_dir is not None
    assert doc_dir.is_symlink


@pytest.mark.parametrize(
    "path_defs,expected_size",
    [
        (
            [],
            1,
        ),
        # Dirs cost 1 a piece
        (
            [virtual_path_def("usr/")],
            2,
        ),
        (
            [virtual_path_def("usr/bin/")],
            3,
        ),
        # Empty file cost 0
        (
            [virtual_path_def("usr/bin/foo", content="")],
            3,
        ),
        # Non-empty files cost 1 per "started" kB
        (
            [virtual_path_def("usr/bin/foo", content="A" * 10)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", content="A" * 1023)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", content="A" * 1024)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", content="A" * 1025)],
            5,
        ),
        # Symlinks costs per started kB of link target
        # - in theory, an empty link target would be free, but `ln -s` does not
        #   allow these, so we are not testing it.
        (
            [virtual_path_def("usr/bin/foo", link_target="A" * 10)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", link_target="A" * 1023)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", link_target="A" * 1024)],
            4,
        ),
        (
            [virtual_path_def("usr/bin/foo", link_target="A" * 1025)],
            5,
        ),
    ],
)
def test_compute_installed_size(
    path_defs: Sequence[PathDef], expected_size: int
) -> None:
    fs = build_virtual_fs(path_defs)
    assert compute_installed_size(fs) == expected_size


@pytest.mark.parametrize(
    "package_name,package_arch,ctrl_members,data_members,expected_result",
    [
        # Arch:all defaults to "no" except when it is `-doc`
        (
            "foo",
            "all",
            [],
            [],
            None,
        ),
        (
            "foo-doc",
            "all",
            [],
            [],
            "foreign",
        ),
        # No matter the name, auto-detection for arch:all stops the moment it has a non-doc path
        (
            "foo-doc",
            "all",
            [],
            [virtual_path_def("/usr/bin/")],
            None,
        ),
        # No matter the name, auto-detection for arch:all stops the moment it has maintscript
        (
            "foo-doc",
            "all",
            [virtual_path_def("postinst")],
            [],
            None,
        ),
        # Arch:any defaults to "same" for proper multi-arch'ed paths
        # NB: This test is always run in amd64 context.
        (
            "libfoo1",
            "amd64",
            [],
            [
                virtual_path_def("/usr/lib/x86_64-linux-gnu/libfoo.so.1"),
            ],
            "same",
        ),
        # Allowed documentation that does not disable auto-detection
        (
            "libfoo1",
            "amd64",
            [],
            [
                virtual_path_def("/usr/lib/x86_64-linux-gnu/libfoo.so.1"),
                virtual_path_def("/usr/share/doc/libfoo1/copyright"),
                virtual_path_def("/usr/share/doc/libfoo1/changelog.gz"),
                virtual_path_def("/usr/share/doc/libfoo1/changelog.Debian.amd64.gz"),
                virtual_path_def("/usr/share/doc/libfoo1/NEWS.Debian.gz"),
                virtual_path_def("/usr/share/doc/libfoo1/README.Debian.gz"),
            ],
            "same",
        ),
        # No auto-detection for at non-multi-arched libs
        (
            "libfoo1",
            "amd64",
            [],
            [
                virtual_path_def("/usr/lib/libfoo.so.1"),
            ],
            None,
        ),
        # No auto-detection with unknown docs.
        (
            "libfoo1",
            "amd64",
            [],
            [
                virtual_path_def("/usr/lib/x86_64-linux-gnu/libfoo.so.1"),
                virtual_path_def("/usr/share/doc/libfoo1/custom-doc"),
            ],
            None,
        ),
        # No auto-detection with random paths docs.
        (
            "libfoo1",
            "amd64",
            [],
            [
                virtual_path_def("/usr/bin/"),
                virtual_path_def("/usr/lib/x86_64-linux-gnu/libfoo.so.1"),
            ],
            None,
        ),
    ],
)
def test_auto_compute_multi_arch(
    package_name: str,
    package_arch: str,
    ctrl_members: Sequence[PathDef],
    data_members: Sequence[PathDef],
    expected_result: Optional[str],
    amd64_dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
    dpkg_arch_query: DpkgArchTable,
) -> None:
    pkg = BinaryPackage(
        Deb822(
            {
                "Package": package_name,
                "Architecture": package_arch,
            }
        ),
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        is_main_package=True,
        should_be_acted_on=True,
    )
    ctrl_fs_root = build_virtual_fs(ctrl_members)
    data_fs_root = build_virtual_fs(data_members)

    assert auto_compute_multi_arch(pkg, ctrl_fs_root, data_fs_root) == expected_result


@pytest.mark.parametrize(
    "data_members,conffiles_lines,expected",
    [
        # Non-existent by default
        (
            [],
            [],
            None,
        ),
        (
            # Normally this would be a conffile, but for the sake of the test, we assume it is not.
            [virtual_path_def("etc/foo", content="test")],
            [],
            {"etc/foo": "098f6bcd4621d373cade4e832627b4f6"},
        ),
        (
            [virtual_path_def("etc/foo", content="test")],
            ["/etc/foo"],
            # Conffiles are not included in the md5sums file (not sure why, but that is
            # "how we always did it")
            None,
        ),
        (
            [virtual_path_def("usr/share/data/foo", content="test")],
            # The code should gracefully cope with dpkg conffile instructions
            ["remove-on-upgrade /etc/foo.conf"],
            {"usr/share/data/foo": "098f6bcd4621d373cade4e832627b4f6"},
        ),
    ],
)
def test_generate_md5sums_file(
    data_members: Sequence[PathDef],
    conffiles_lines: Sequence[str],
    expected: Optional[Mapping[str, str]],
) -> None:

    if conffiles_lines:
        conffiles_def = virtual_path_def(
            "conffiles", content="\n".join(conffiles_lines)
        )
        ctrl_fs = build_virtual_fs([conffiles_def], read_write_fs=True)
    else:
        ctrl_fs = build_virtual_fs([], read_write_fs=True)
    data_fs = build_virtual_fs(data_members)

    generate_md5sums_file(
        ctrl_fs,
        data_fs,
    )
    md5sums_file = ctrl_fs.get("md5sums")
    if expected is None:
        assert md5sums_file is None
        return

    assert md5sums_file is not None

    actual = {}
    for line in md5sums_file.open():
        checksum, fs_path = line.split()
        actual[fs_path] = checksum

    assert actual == expected



import asyncio
import dataclasses
import subprocess
import sys
from collections import defaultdict
from typing import Literal, Optional, Sequence, Iterable, Mapping

from debian.deb822 import Deb822
from debian.debian_support import Version

AptCacheState = Literal[
    "not-loaded",
    "loading",
    "loaded",
    "failed",
    "tooling-not-available",
    "empty-cache",
]


@dataclasses.dataclass(slots=True)
class PackageInformation:
    name: str
    architecture: str
    version: Version
    multi_arch: str
    # suites: Sequence[Tuple[str, ...]]
    synopsis: str
    section: str
    provides: Optional[str]
    upstream_homepage: Optional[str]


@dataclasses.dataclass(slots=True, frozen=True)
class PackageLookup:
    name: str
    package: Optional[PackageInformation]
    provided_by: Sequence[PackageInformation]


class AptCache:

    def __init__(self) -> None:
        self._state: AptCacheState = "not-loaded"
        self._load_error: Optional[str] = None
        self._lookups: Mapping[str, PackageLookup] = {}

    @property
    def state(self) -> AptCacheState:
        return self._state

    @property
    def load_error(self) -> Optional[str]:
        return self._load_error

    def lookup(self, name: str) -> Optional[PackageLookup]:
        return self._lookups.get(name)

    async def load(self) -> None:
        if self._state in ("loading", "loaded"):
            raise RuntimeError(f"Already {self._state}")
        self._load_error = None
        self._state = "loading"
        try:
            files_raw = subprocess.check_output(
                [
                    "apt-get",
                    "indextargets",
                    "--format",
                    "$(IDENTIFIER)\x1f$(FILENAME)",
                ]
            ).decode("utf-8")
        except FileNotFoundError:
            self._state = "tooling-not-available"
            self._load_error = "apt-get not available in PATH"
            return
        except subprocess.CalledProcessError as e:
            self._state = "failed"
            self._load_error = f"apt-get exited with {e.returncode}"
            return
        packages = {}
        for raw_file_line in files_raw.split("\n"):
            if not raw_file_line or raw_file_line.isspace():
                continue
            identifier, filename = raw_file_line.split("\x1f")
            if identifier not in ("Packages",):
                continue
            try:
                for package_info in parse_apt_file(filename):
                    # Let other computations happen if needed.
                    await asyncio.sleep(0)
                    existing = packages.get(package_info.name)
                    if existing and package_info.version < existing.version:
                        continue
                    packages[package_info.name] = package_info
            except FileNotFoundError:
                self._state = "tooling-not-available"
                self._load_error = "/usr/lib/apt/apt-helper not available"
                return
            except (AttributeError, RuntimeError, IndexError) as e:
                self._state = "failed"
                self._load_error = str(e)
                return
        provides = defaultdict(list)
        for package_info in packages.values():
            if not package_info.provides:
                continue
            # Some packages (`debhelper`) provides the same package multiple times (`debhelper-compat`).
            # Normalize that into one.
            deps = {
                clause.split("(")[0].strip()
                for clause in package_info.provides.split(",")
            }
            for dep in sorted(deps):
                provides[dep].append(package_info)

        self._lookups = {
            name: PackageLookup(
                name,
                packages.get(name),
                tuple(provides.get(name, [])),
            )
            for name in packages.keys() | provides.keys()
        }
        self._state = "loaded"


def parse_apt_file(filename: str) -> Iterable[PackageInformation]:
    proc = subprocess.Popen(
        ["/usr/lib/apt/apt-helper", "cat-file", filename],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
    )
    with proc:
        for stanza in Deb822.iter_paragraphs(proc.stdout):
            pkg_info = stanza_to_package_info(stanza)
            if pkg_info is not None:
                yield pkg_info


def stanza_to_package_info(stanza: Deb822) -> Optional[PackageInformation]:
    try:
        name = stanza["Package"]
        architecture = sys.intern(stanza["Architecture"])
        version = Version(stanza["Version"])
        multi_arch = sys.intern(stanza.get("Multi-Arch", "no"))
        synopsis = stanza["Description"]
        section = sys.intern(stanza["Section"])
        provides = stanza.get("Provides")
        homepage = stanza.get("Homepage")
    except KeyError:
        return None
    if "\n" in synopsis:
        # "Modern" Packages files do not have the full description. But in case we see a (very old one)
        # have consistent behavior with the modern ones.
        synopsis = synopsis.split("\n")[0]

    return PackageInformation(
        name,
        architecture,
        version,
        multi_arch,
        synopsis,
        section,
        provides,
        homepage,
    )



from pathlib import Path

import pytest

from debputy.util import compute_output_filename


def write_unpacked_deb(root: Path, package: str, version: str, arch: str):
    (root / "control").write_text(
        f"Package: {package}\nVersion: {version}\nArchitecture: {arch}\n"
    )


@pytest.mark.parametrize(
    "package,version,arch,is_udeb,expected",
    [
        ("fake", "1.0", "amd64", False, "fake_1.0_amd64.deb"),
        ("fake", "1.0", "amd64", True, "fake_1.0_amd64.udeb"),
        ("fake", "2:1.0", "amd64", False, "fake_1.0_amd64.deb"),
        ("fake", "2:1.0", "amd64", True, "fake_1.0_amd64.udeb"),
        ("fake", "3.0", "all", False, "fake_3.0_all.deb"),
        ("fake", "3.0", "all", True, "fake_3.0_all.udeb"),
    ],
)
def test_generate_deb_filename(tmp_path, package, version, arch, is_udeb, expected):
    write_unpacked_deb(tmp_path, package, version, arch)
    assert compute_output_filename(str(tmp_path), is_udeb) == expected



import pytest

from debputy.architecture_support import faked_arch_table
from debputy.commands.debputy_cmd.output import no_fancy_output
from debputy.dh_migration.models import (
    DHMigrationSubstitution,
    AcceptableMigrationIssues,
    FeatureMigration,
)
from debputy.filesystem_scan import FSROOverlay
from debputy.highlevel_manifest import MutableYAMLManifest
from debputy.substitution import SubstitutionImpl, VariableContext

MOCK_ENV = {
    # This conflicts with the dpkg arch table intentionally (to ensure we can tell which one is being resolved)
    "DEB_HOST_ARCHITECTURE": "i386",
}
MOCK_DPKG_ARCH_TABLE = faked_arch_table("amd64", build_arch="i386")
MOCK_VARIABLE_CONTEXT = VariableContext(FSROOverlay.create_root_dir("debian", "debian"))


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "unchanged",
            "unchanged",
        ),
        (
            "unchanged\\{{\n}}",
            "unchanged\\{{\n}}",
        ),  # Newline is not an allowed part of a substitution
        (
            "{{token:DOUBLE_OPEN_CURLY_BRACE}}{{token:NL}}{{token:DOUBLE_CLOSE_CURLY_BRACE}}",
            "{{\n}}",
        ),
        (
            "{{token:DOUBLE_OPEN_CURLY_BRACE}}token:TAB}}{{token:TAB{{token:DOUBLE_CLOSE_CURLY_BRACE}}",
            "{{token:TAB}}{{token:TAB}}",
        ),
        (
            "/usr/lib/{{DEB_HOST_MULTIARCH}}",
            f'/usr/lib/{MOCK_DPKG_ARCH_TABLE["DEB_HOST_MULTIARCH"]}',
        ),
    ],
)
def test_substitution_match(debputy_plugin_feature_set, value, expected) -> None:
    subst = SubstitutionImpl(
        plugin_feature_set=debputy_plugin_feature_set,
        dpkg_arch_table=MOCK_DPKG_ARCH_TABLE,
        environment=MOCK_ENV,
        variable_context=MOCK_VARIABLE_CONTEXT,
    )
    replacement = subst.substitute(value, "test def")
    assert replacement == expected


def test_migrate_substitution() -> None:
    feature_migration = FeatureMigration("test migration", no_fancy_output())
    subst = DHMigrationSubstitution(
        MOCK_DPKG_ARCH_TABLE,
        AcceptableMigrationIssues(frozenset()),
        feature_migration,
        MutableYAMLManifest({}),
    )
    replacement = subst.substitute("usr/lib/${DEB_HOST_MULTIARCH}/foo", "test def")
    assert replacement == "usr/lib/{{DEB_HOST_MULTIARCH}}/foo"



#!/usr/bin/python3
import os
import subprocess
import sys
from typing import Optional

SOURCE_ROOT = os.path.dirname(os.path.dirname(__file__))

inc_path = os.path.join(SOURCE_ROOT, "src")
sys.path.insert(0, inc_path)

from debputy.util import _error

from debputy.lsp.lsp_reference_keyword import LSP_DATA_DOMAIN

from debputy.lsp.lsp_debian_control import (
    substvars_metadata,
    substvars_metadata_basename,
)
from debputy.lsp.lsp_debian_control_reference_data import (
    TRANSLATABLE_DEB822_FILE_METADATA,
    lsp_reference_data_dir,
    Deb822KnownField,
)

try:
    from polib import pofile, POEntry, POFile
except ImportError:
    _error("This script needs polib (python3-polib)")


def strip_source_root(filename: str) -> str:
    if filename.startswith(SOURCE_ROOT + "/"):
        filename = filename[len(SOURCE_ROOT) + 1 :]
    return filename


def pot_file_from_template(raw_template: str) -> POFile:
    pot_file = pofile(raw_template, encoding="utf-8")
    pot_file.clear()
    pot_file.metadata["Content-Type"] = "text/plain; charset=utf-8"
    return pot_file


def generate_po_from_docs(
    pot_file: POFile,
    filename: str,
    msgctxt: str,
    topic: str,
    synopsis: str,
    hover_doc: str,
) -> None:
    if synopsis:
        pot_file.append(
            POEntry(
                msgctxt=msgctxt,
                msgid=synopsis,
                occurrences=[(filename, 0)],
                comment=f"[Synopsis] One-line description of {topic} [Plaintext]. Shown with completion (etc.)",
                flags=[],
            )
        )
    if hover_doc and hover_doc != synopsis:
        pot_file.append(
            POEntry(
                msgctxt=msgctxt,
                msgid=hover_doc,
                occurrences=[(filename, 0)],
                comment=f"[Hover Doc] Extended description of {topic} [Markdown]. Shown in hover docs (etc.)",
                flags=["python-brace-format"],
            )
        )


def generate_lsp_data_pot(raw_template: str) -> POFile:
    pot_file = pot_file_from_template(raw_template)

    for file_metadata_constructor in TRANSLATABLE_DEB822_FILE_METADATA:
        file_metadata = file_metadata_constructor()
        name = file_metadata.reference_data_basename
        filename = strip_source_root(os.path.join(lsp_reference_data_dir(), name))

        for stanza in file_metadata.stanza_types():
            field: Deb822KnownField
            for field in stanza.stanza_fields.values():
                field_synopsis = field.synopsis
                field_hover_doc = field.long_description
                known_values = field.known_values

                generate_po_from_docs(
                    pot_file,
                    filename,
                    f"Stanza:{stanza.stanza_type_name}|Field:{field.name}",
                    "the field itself",
                    field_synopsis,
                    field_hover_doc,
                )

                if known_values:
                    for keyword in known_values.values():
                        kw_synopsis = keyword.synopsis
                        kw_hover_doc = keyword.long_description

                        generate_po_from_docs(
                            pot_file,
                            filename,
                            f"Stanza:{stanza.stanza_type_name}|Field:{field.name}",
                            f'the value "{keyword.value}"',
                            kw_synopsis,
                            kw_hover_doc,
                        )

    substvars_metadata_file = strip_source_root(
        os.path.join(lsp_reference_data_dir(), substvars_metadata_basename())
    )
    for substvar in substvars_metadata().values():
        generate_po_from_docs(
            pot_file,
            substvars_metadata_file,
            f"Substvars:{substvar.name}",
            f'the substvar "{substvar.name}"',
            substvar.synopsis,
            substvar.description,
        )

    return pot_file


def main() -> None:
    os.environ["TZ"] = "UTC"
    output_dir = os.path.join(SOURCE_ROOT, "po")

    try:
        debputy_version = (
            subprocess.check_output(
                [os.path.join(SOURCE_ROOT, "debputy.sh"), "--version"]
            )
            .decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError:
        debputy_version = "VERSION"

    # We are just doing this for the header (metadata entry)
    raw_template = subprocess.check_output(
        [
            "xgettext",
            "-L",
            "Python",
            "--package-name=debputy",
            f"--package-version={debputy_version}",
            "--from-code=utf-8",
            "--foreign-user",
            "--msgid-bugs-address=https://salsa.debian.org/debian/debputy/-/issues/",
            "/dev/stdin",
            "-o-",
        ],
        input=b'_("foo")',
    ).decode("utf-8")
    lsp_data_pot = generate_lsp_data_pot(raw_template)

    output_name = os.path.join(output_dir, LSP_DATA_DOMAIN, "messages.pot")

    with open(output_name, "w", encoding="utf-8") as fd:
        fd.write(str(lsp_data_pot))
    print(f"Generated {output_name}")


if __name__ == "__main__":
    main()



from .exceptions import InvalidStyle as InvalidStyle
from .library import Library as Library
from .utilities import Utilities as Utilities

class MetaStyle(type):
    def __getattr__(cls, color: str): ...

class Style(metaclass=MetaStyle):
    @classmethod
    def underline_color(cls, color: str | int) -> str: ...
    @classmethod
    def UNDERLINE_COLOR(cls, color: str | int) -> str: ...

class style(Style): ...



import argparse
import contextlib
import os
import re
import shutil
import subprocess
import sys
from typing import (
    Union,
    Sequence,
    Iterator,
    IO,
    Mapping,
    Tuple,
    Optional,
    Any,
)

from debputy.util import assume_not_none

try:
    import colored

    if (
        not hasattr(colored, "Style")
        or not hasattr(colored, "Fore")
        or not hasattr(colored, "Back")
    ):
        # Seen with python3-colored v1 (bookworm)
        raise ImportError
except ImportError:
    colored = None


def _pager() -> Optional[str]:
    pager = os.environ.get("DEBPUTY_PAGER")
    if pager is None:
        pager = os.environ.get("PAGER")
    if pager is None and shutil.which("less") is not None:
        pager = "less"
    return pager


URL_START = "\033]8;;"
URL_END = "\033]8;;\a"
MAN_URL_REWRITE = re.compile(r"man:(\S+)[(](\d+)[)]")

_SUPPORTED_COLORS = {
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
}
_SUPPORTED_STYLES = {"none", "bold"}


class OutputStylingBase:
    def __init__(
        self,
        stream: IO[str],
        output_format: str,
        *,
        optimize_for_screen_reader: bool = False,
    ) -> None:
        self.stream = stream
        self.output_format = output_format
        self.optimize_for_screen_reader = optimize_for_screen_reader
        self._color_support = None

    def colored(
        self,
        text: str,
        *,
        fg: Optional[Union[str]] = None,
        bg: Optional[str] = None,
        style: Optional[str] = None,
    ) -> str:
        self._check_color(fg)
        self._check_color(bg)
        self._check_text_style(style)
        return text

    @property
    def supports_colors(self) -> bool:
        return False

    def print_list_table(
        self,
        headers: Sequence[Union[str, Tuple[str, str]]],
        rows: Sequence[Sequence[str]],
    ) -> None:
        if rows:
            if any(len(r) != len(rows[0]) for r in rows):
                raise ValueError(
                    "Unbalanced table: All rows must have the same column count"
                )
            if len(rows[0]) != len(headers):
                raise ValueError(
                    "Unbalanced table: header list does not agree with row list on number of columns"
                )

        if not headers:
            raise ValueError("No headers provided!?")

        cadjust = {}
        header_names = []
        for c in headers:
            if isinstance(c, str):
                header_names.append(c)
            else:
                cname, adjust = c
                header_names.append(cname)
                cadjust[cname] = adjust

        if self.output_format == "csv":
            from csv import writer

            w = writer(self.stream)
            w.writerow(header_names)
            w.writerows(rows)
            return

        column_lengths = [
            max((len(h), max(len(r[i]) for r in rows)))
            for i, h in enumerate(header_names)
        ]
        # divider => "+---+---+-...-+"
        divider = "+-" + "-+-".join("-" * x for x in column_lengths) + "-+"
        # row_format => '| {:<10} | {:<8} | ... |' where the numbers are the column lengths
        row_format_inner = " | ".join(
            f"{{CELL_COLOR}}{{:{cadjust.get(cn, '<')}{x}}}{{CELL_COLOR_RESET}}"
            for cn, x in zip(header_names, column_lengths)
        )

        row_format = f"| {row_format_inner} |"

        if self.supports_colors:
            cs = self._color_support
            assert cs is not None
            header_color = cs.Style.bold
            header_color_reset = cs.Style.reset
        else:
            header_color = ""
            header_color_reset = ""

        self.print_visual_formatting(divider)
        self.print(
            row_format.format(
                *header_names,
                CELL_COLOR=header_color,
                CELL_COLOR_RESET=header_color_reset,
            )
        )
        self.print_visual_formatting(divider)
        for row in rows:
            self.print(row_format.format(*row, CELL_COLOR="", CELL_COLOR_RESET=""))
        self.print_visual_formatting(divider)

    def print(self, /, string: str = "", **kwargs) -> None:
        if "file" in kwargs:
            raise ValueError("Unsupported kwarg file")
        print(string, file=self.stream, **kwargs)

    def print_visual_formatting(self, /, format_sequence: str, **kwargs) -> None:
        if self.optimize_for_screen_reader:
            return
        self.print(format_sequence, **kwargs)

    def print_for_screen_reader(self, /, text: str, **kwargs) -> None:
        if not self.optimize_for_screen_reader:
            return
        self.print(text, **kwargs)

    def _check_color(self, color: Optional[str]) -> None:
        if color is not None and color not in _SUPPORTED_COLORS:
            raise ValueError(
                f"Unsupported color: {color}. Only the following are supported {','.join(_SUPPORTED_COLORS)}"
            )

    def _check_text_style(self, style: Optional[str]) -> None:
        if style is not None and style not in _SUPPORTED_STYLES:
            raise ValueError(
                f"Unsupported style: {style}. Only the following are supported {','.join(_SUPPORTED_STYLES)}"
            )

    def render_url(self, link_url: str) -> str:
        return link_url

    def bts(self, bugno) -> str:
        return f"https://bugs.debian.org/{bugno}"


class ANSIOutputStylingBase(OutputStylingBase):
    def __init__(
        self,
        stream: IO[str],
        output_format: str,
        *,
        support_colors: bool = True,
        support_clickable_urls: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(stream, output_format, **kwargs)
        self._stream = stream
        self._color_support = colored
        self._support_colors = (
            support_colors if self._color_support is not None else False
        )
        self._support_clickable_urls = support_clickable_urls

    @property
    def supports_colors(self) -> bool:
        return self._support_colors

    def colored(
        self,
        text: str,
        *,
        fg: Optional[str] = None,
        bg: Optional[str] = None,
        style: Optional[str] = None,
    ) -> str:
        self._check_color(fg)
        self._check_color(bg)
        self._check_text_style(style)
        _colored = self._color_support
        if not self.supports_colors or _colored is None:
            return text
        codes = []
        if style is not None:
            code = getattr(_colored.Style, style)
            assert code is not None
            codes.append(code)
        if fg is not None:
            code = getattr(_colored.Fore, fg)
            assert code is not None
            codes.append(code)
        if bg is not None:
            code = getattr(_colored.Back, bg)
            assert code is not None
            codes.append(code)
        if not codes:
            return text
        return "".join(codes) + text + _colored.Style.reset

    def render_url(self, link_url: str) -> str:
        if not self._support_clickable_urls:
            return super().render_url(link_url)
        link_text = link_url
        if not self.optimize_for_screen_reader and link_url.startswith("man:"):
            # Rewrite man page to a clickable link by default. I am not sure how the hyperlink
            # ANSI code works with screen readers, so lets not rewrite the man page link by
            # default. My fear is that both the link url and the link text gets read out.
            m = MAN_URL_REWRITE.match(link_url)
            if m:
                page, section = m.groups()
                link_url = f"https://manpages.debian.org/{page}.{section}"
        return URL_START + f"{link_url}\a{link_text}" + URL_END

    def bts(self, bugno) -> str:
        if not self._support_clickable_urls:
            return super().bts(bugno)
        return self.render_url(f"https://bugs.debian.org/{bugno}")


def no_fancy_output(
    stream: IO[str] = None,
    output_format: str = str,
    optimize_for_screen_reader: bool = False,
) -> OutputStylingBase:
    if stream is None:
        stream = sys.stdout
    return OutputStylingBase(
        stream,
        output_format,
        optimize_for_screen_reader=optimize_for_screen_reader,
    )


def _output_styling(
    parsed_args: argparse.Namespace,
    stream: IO[str],
) -> OutputStylingBase:
    output_format = getattr(parsed_args, "output_format", None)
    if output_format is None:
        output_format = "text"
    optimize_for_screen_reader = os.environ.get("OPTIMIZE_FOR_SCREEN_READER", "") != ""
    if not stream.isatty():
        return no_fancy_output(
            stream,
            output_format,
            optimize_for_screen_reader=optimize_for_screen_reader,
        )

    return ANSIOutputStylingBase(
        stream, output_format, optimize_for_screen_reader=optimize_for_screen_reader
    )


@contextlib.contextmanager
def _stream_to_pager(
    parsed_args: argparse.Namespace,
) -> Iterator[Tuple[IO[str], OutputStylingBase]]:
    fancy_output = _output_styling(parsed_args, sys.stdout)
    if (
        not parsed_args.pager
        or not sys.stdout.isatty()
        or fancy_output.output_format != "text"
    ):
        yield sys.stdout, fancy_output
        return

    pager = _pager()
    if pager is None:
        yield sys.stdout, fancy_output
        return

    env: Mapping[str, str] = os.environ
    if "LESS" not in env:
        env_copy = dict(os.environ)
        env_copy["LESS"] = "-FRSXMQ"
        env = env_copy

    cmd = subprocess.Popen(
        pager,
        stdin=subprocess.PIPE,
        encoding="utf-8",
        env=env,
    )
    stdin = assume_not_none(cmd.stdin)
    try:
        fancy_output.stream = stdin
        yield stdin, fancy_output
    except Exception:
        stdin.close()
        cmd.kill()
        cmd.wait()
        raise
    finally:
        fancy_output.stream = sys.stdin
    stdin.close()
    cmd.wait()



import functools
import os
import re
import subprocess
from typing import Any, Optional

from debputy.plugin.api import (
    DebputyPluginInitializer,
    BinaryCtrlAccessor,
    PackageProcessingContext,
    VirtualPath,
)
from debputy.util import _error


_RE_GRANTLEE_VERSION = re.compile(r"^\d\.\d$")


def initialize(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector(
        "detect-grantlee-dependencies",
        detect_grantlee_dependencies,
    )


def detect_grantlee_dependencies(
    fs_root: VirtualPath,
    ctrl: BinaryCtrlAccessor,
    context: PackageProcessingContext,
) -> None:
    binary_package = context.binary_package
    if binary_package.is_arch_all:
        # Delta from dh_grantlee, but the MULTIARCH paths should not
        # exist in arch:all packages
        return
    ma = binary_package.package_deb_architecture_variable("MULTIARCH")
    grantlee_root_dirs = [
        f"usr/lib/{ma}/grantlee",
        f"usr/lib/{ma}/qt5/plugins/grantlee",
    ]
    grantee_version: Optional[str] = None
    for grantlee_root_dir in grantlee_root_dirs:
        grantlee_root_path = fs_root.lookup(grantlee_root_dir)
        if grantlee_root_path is None or not grantlee_root_path.is_dir:
            continue
        # Delta: The original code recurses and then checks for "grantee/VERSION".
        # Code here assumes that was just File::Find being used as a dir iterator.
        for child in grantlee_root_path.iterdir:
            if not _RE_GRANTLEE_VERSION.fullmatch(child.name):
                continue
            version = child.name
            if grantee_version is not None and version != grantee_version:
                _error(
                    f"Package {binary_package.name} contains plugins for different grantlee versions"
                )
            grantee_version = version

    if grantee_version is None:
        return
    dep_version = grantee_version.replace(".", "-")
    grantee_dep = f"grantlee5-templates-{dep_version}"
    ctrl.substvars["grantlee:Depends"] = grantee_dep



import dataclasses
from enum import Enum
from typing import List, Callable, Optional, Sequence, Any, Self, Mapping

from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.packages import BinaryPackage
from debputy.substitution import Substitution
from debputy.util import active_profiles_match


@dataclasses.dataclass(slots=True, frozen=True)
class ConditionContext:
    binary_package: Optional[BinaryPackage]
    deb_options_and_profiles: DebBuildOptionsAndProfiles
    substitution: Substitution
    dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable
    dpkg_arch_query_table: DpkgArchTable

    def replace(self, /, **changes: Any) -> "Self":
        return dataclasses.replace(self, **changes)


class ManifestCondition(DebputyDispatchableType):
    __slots__ = ()

    def describe(self) -> str:
        raise NotImplementedError

    def negated(self) -> "ManifestCondition":
        return NegatedManifestCondition(self)

    def evaluate(self, context: ConditionContext) -> bool:
        raise NotImplementedError

    @classmethod
    def _manifest_group(
        cls,
        match_type: "_ConditionGroupMatchType",
        conditions: "Sequence[ManifestCondition]",
    ) -> "ManifestCondition":
        condition = conditions[0]
        if (
            isinstance(condition, ManifestConditionGroup)
            and condition.match_type == match_type
        ):
            return condition.extend(conditions[1:])
        return ManifestConditionGroup(match_type, conditions)

    @classmethod
    def any_of(cls, conditions: "Sequence[ManifestCondition]") -> "ManifestCondition":
        return cls._manifest_group(_ConditionGroupMatchType.ANY_OF, conditions)

    @classmethod
    def all_of(cls, conditions: "Sequence[ManifestCondition]") -> "ManifestCondition":
        return cls._manifest_group(_ConditionGroupMatchType.ALL_OF, conditions)

    @classmethod
    def is_cross_building(cls) -> "ManifestCondition":
        return _IS_CROSS_BUILDING

    @classmethod
    def can_execute_compiled_binaries(cls) -> "ManifestCondition":
        return _CAN_EXECUTE_COMPILED_BINARIES

    @classmethod
    def run_build_time_tests(cls) -> "ManifestCondition":
        return _RUN_BUILD_TIME_TESTS


class NegatedManifestCondition(ManifestCondition):
    __slots__ = ("_condition",)

    def __init__(self, condition: ManifestCondition) -> None:
        super().__init__()
        self._condition = condition

    def negated(self) -> "ManifestCondition":
        return self._condition

    def describe(self) -> str:
        return f"not ({self._condition.describe()})"

    def evaluate(self, context: ConditionContext) -> bool:
        return not self._condition.evaluate(context)


class _ConditionGroupMatchType(Enum):
    ANY_OF = (any, "At least one of: [{conditions}]")
    ALL_OF = (all, "All of: [{conditions}]")

    def describe(self, conditions: Sequence[ManifestCondition]) -> str:
        return self.value[1].format(
            conditions=", ".join(x.describe() for x in conditions)
        )

    def evaluate(
        self, conditions: Sequence[ManifestCondition], context: ConditionContext
    ) -> bool:
        return self.value[0](c.evaluate(context) for c in conditions)


class ManifestConditionGroup(ManifestCondition):
    __slots__ = ("match_type", "_conditions")

    def __init__(
        self,
        match_type: _ConditionGroupMatchType,
        conditions: Sequence[ManifestCondition],
    ) -> None:
        super().__init__()
        self.match_type = match_type
        self._conditions = conditions

    def describe(self) -> str:
        return self.match_type.describe(self._conditions)

    def evaluate(self, context: ConditionContext) -> bool:
        return self.match_type.evaluate(self._conditions, context)

    def extend(
        self,
        conditions: Sequence[ManifestCondition],
    ) -> "ManifestConditionGroup":
        combined = list(self._conditions)
        combined.extend(conditions)
        return ManifestConditionGroup(
            self.match_type,
            combined,
        )


class ArchMatchManifestConditionBase(ManifestCondition):
    __slots__ = ("_arch_spec", "_is_negated")

    def __init__(self, arch_spec: List[str], *, is_negated: bool = False) -> None:
        super().__init__()
        self._arch_spec = arch_spec
        self._is_negated = is_negated

    def negated(self) -> "ManifestCondition":
        return self.__class__(self._arch_spec, is_negated=not self._is_negated)


class SourceContextArchMatchManifestCondition(ArchMatchManifestConditionBase):
    def describe(self) -> str:
        if self._is_negated:
            return f'architecture (for source package) matches *none* of [{", ".join(self._arch_spec)}]'
        return f'architecture (for source package) matches any of [{", ".join(self._arch_spec)}]'

    def evaluate(self, context: ConditionContext) -> bool:
        arch = context.dpkg_architecture_variables.current_host_arch
        match = context.dpkg_arch_query_table.architecture_is_concerned(
            arch, self._arch_spec
        )
        return not match if self._is_negated else match


class BinaryPackageContextArchMatchManifestCondition(ArchMatchManifestConditionBase):
    def describe(self) -> str:
        if self._is_negated:
            return f'architecture (for binary package) matches *none* of [{", ".join(self._arch_spec)}]'
        return f'architecture (for binary package) matches any of [{", ".join(self._arch_spec)}]'

    def evaluate(self, context: ConditionContext) -> bool:
        binary_package = context.binary_package
        if binary_package is None:
            raise RuntimeError(
                "Condition only applies in the context of a BinaryPackage, but was evaluated"
                " without one"
            )
        arch = binary_package.resolved_architecture
        match = context.dpkg_arch_query_table.architecture_is_concerned(
            arch, self._arch_spec
        )
        return not match if self._is_negated else match


class BuildProfileMatch(ManifestCondition):
    __slots__ = ("_profile_spec", "_is_negated")

    def __init__(self, profile_spec: str, *, is_negated: bool = False) -> None:
        super().__init__()
        self._profile_spec = profile_spec
        self._is_negated = is_negated

    def negated(self) -> "ManifestCondition":
        return self.__class__(self._profile_spec, is_negated=not self._is_negated)

    def describe(self) -> str:
        if self._is_negated:
            return f"DEB_BUILD_PROFILES matches *none* of [{self._profile_spec}]"
        return f"DEB_BUILD_PROFILES matches any of [{self._profile_spec}]"

    def evaluate(self, context: ConditionContext) -> bool:
        match = active_profiles_match(
            self._profile_spec, context.deb_options_and_profiles.deb_build_profiles
        )
        return not match if self._is_negated else match


@dataclasses.dataclass(frozen=True, slots=True)
class _SingletonCondition(ManifestCondition):
    description: str
    implementation: Callable[[ConditionContext], bool]

    def describe(self) -> str:
        return self.description

    def evaluate(self, context: ConditionContext) -> bool:
        return self.implementation(context)


def _can_run_built_binaries(context: ConditionContext) -> bool:
    if not context.dpkg_architecture_variables.is_cross_compiling:
        return True
    # User / Builder asserted that we could even though we are cross-compiling, so we have to assume it is true
    return (
        "crossbuildcanrunhostbinaries"
        in context.deb_options_and_profiles.deb_build_options
    )


def _run_build_time_tests(deb_build_options: Mapping[str, Optional[str]]) -> bool:
    return "nocheck" not in deb_build_options


_IS_CROSS_BUILDING = _SingletonCondition(
    "Cross Compiling (i.e., DEB_HOST_GNU_TYPE != DEB_BUILD_GNU_TYPE)",
    lambda c: c.dpkg_architecture_variables.is_cross_compiling,
)

_CAN_EXECUTE_COMPILED_BINARIES = _SingletonCondition(
    "Can run built binaries (natively or via transparent emulation)",
    _can_run_built_binaries,
)

_RUN_BUILD_TIME_TESTS = _SingletonCondition(
    "Run build time tests",
    lambda c: _run_build_time_tests(c.deb_options_and_profiles.deb_build_options),
)

_BUILD_DOCS_BDO = _SingletonCondition(
    "Build docs (nodocs not in DEB_BUILD_OPTIONS)",
    lambda c: "nodocs" not in c.deb_options_and_profiles.deb_build_options,
)


del _SingletonCondition
del _can_run_built_binaries



from typing import List

from debputy import elf_util
from debputy.elf_util import ELF_LINKING_TYPE_DYNAMIC
from debputy.plugin.api import (
    VirtualPath,
    PackageProcessingContext,
)
from debputy.plugin.api.impl import BinaryCtrlAccessorProvider

SKIPPED_DEBUG_DIRS = [
    "lib",
    "lib64",
    "usr",
    "bin",
    "sbin",
    "opt",
    "dev",
    "emul",
    ".build-id",
]

SKIP_DIRS = {f"./usr/lib/debug/{subdir}" for subdir in SKIPPED_DEBUG_DIRS}


def _walk_filter(fs_path: VirtualPath, children: List[VirtualPath]) -> bool:
    if fs_path.path in SKIP_DIRS:
        children.clear()
        return False
    return True


def detect_shlibdeps(
    fs_root: VirtualPath,
    ctrl: BinaryCtrlAccessorProvider,
    _context: PackageProcessingContext,
) -> None:
    elf_files_to_process = elf_util.find_all_elf_files(
        fs_root,
        walk_filter=_walk_filter,
        with_linking_type=ELF_LINKING_TYPE_DYNAMIC,
    )

    if not elf_files_to_process:
        return

    ctrl.dpkg_shlibdeps(elf_files_to_process)



import os
from typing import Mapping

import pytest
from debian.deb822 import Deb822
from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import (
    DpkgArchitectureBuildProcessValuesTable,
    faked_arch_table,
)
from debputy.filesystem_scan import FSROOverlay
from debputy.manifest_parser.util import AttributePath
from debputy.packages import BinaryPackage, SourcePackage
from debputy.plugin.debputy.debputy_plugin import initialize_debputy_features
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
)
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.impl import (
    DebputyPluginInitializerProvider,
    _resolve_bundled_plugin_docs_path,
)
from debputy.substitution import (
    NULL_SUBSTITUTION,
    Substitution,
    SubstitutionImpl,
    VariableContext,
)

# Disable dpkg's translation layer.  It is very slow and disabling it makes it easier to debug
# test-failure reports from systems with translations active.
os.environ["DPKG_NLS"] = "0"


@pytest.fixture(scope="session")
def amd64_dpkg_architecture_variables() -> DpkgArchitectureBuildProcessValuesTable:
    return faked_arch_table("amd64")


@pytest.fixture(scope="session")
def dpkg_arch_query() -> DpkgArchTable:
    return DpkgArchTable.load_arch_table()


@pytest.fixture()
def source_package() -> SourcePackage:
    return SourcePackage(
        {
            "Source": "foo",
        }
    )


@pytest.fixture()
def package_single_foo_arch_all_cxt_amd64(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
) -> Mapping[str, BinaryPackage]:
    return {
        p.name: p
        for p in [
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo",
                        "Architecture": "all",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
                is_main_package=True,
            )
        ]
    }


@pytest.fixture()
def package_foo_w_udeb_arch_any_cxt_amd64(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
) -> Mapping[str, BinaryPackage]:
    return {
        p.name: p
        for p in [
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo",
                        "Architecture": "any",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
                is_main_package=True,
            ),
            BinaryPackage(
                Deb822(
                    {
                        "Package": "foo-udeb",
                        "Architecture": "any",
                        "Package-Type": "udeb",
                    }
                ),
                amd64_dpkg_architecture_variables,
                dpkg_arch_query,
            ),
        ]
    }


@pytest.fixture(scope="session")
def null_substitution() -> Substitution:
    return NULL_SUBSTITUTION


@pytest.fixture(scope="session")
def _empty_debputy_plugin_feature_set() -> PluginProvidedFeatureSet:
    return PluginProvidedFeatureSet()


@pytest.fixture(scope="session")
def amd64_substitution(
    amd64_dpkg_architecture_variables,
    _empty_debputy_plugin_feature_set,
) -> Substitution:
    debian_dir = FSROOverlay.create_root_dir("debian", "debian")
    variable_context = VariableContext(
        debian_dir,
    )
    return SubstitutionImpl(
        plugin_feature_set=_empty_debputy_plugin_feature_set,
        dpkg_arch_table=amd64_dpkg_architecture_variables,
        static_variables=None,
        environment={},
        unresolvable_substitutions=frozenset(["SOURCE_DATE_EPOCH", "PACKAGE"]),
        variable_context=variable_context,
    )


@pytest.fixture(scope="session")
def no_profiles_or_build_options() -> DebBuildOptionsAndProfiles:
    return DebBuildOptionsAndProfiles(environ={})


@pytest.fixture(scope="session")
def debputy_plugin_feature_set(
    _empty_debputy_plugin_feature_set, amd64_substitution
) -> PluginProvidedFeatureSet:
    plugin_name = "debputy"
    plugin_metadata = DebputyPluginMetadata(
        plugin_name=plugin_name,
        api_compat_version=1,
        plugin_initializer=initialize_debputy_features,
        plugin_doc_path_resolver=lambda: _resolve_bundled_plugin_docs_path(
            plugin_name,
            initialize_debputy_features,
        ),
        plugin_loader=None,
        plugin_path="<loaded-via-test>",
    )
    feature_set = _empty_debputy_plugin_feature_set
    api = DebputyPluginInitializerProvider(
        plugin_metadata,
        feature_set,
        amd64_substitution,
    )
    api.load_plugin()
    return feature_set


@pytest.fixture
def attribute_path(request) -> AttributePath:
    return AttributePath.builtin_path()[request.node.nodeid]



from typing import (
    Union,
    Any,
    Optional,
    List,
    Tuple,
    Iterable,
    TYPE_CHECKING,
    Callable,
    Sequence,
    get_origin,
    Literal,
    get_args,
)

from debputy.linting.lint_util import LintState
from debputy.lsp.diagnostics import LintSeverity
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debputy.lsp.vendoring._deb822_repro.locatable import (
    Position as TEPosition,
    Range as TERange,
)
from debputy.manifest_parser.declarative_parser import (
    DeclarativeMappingInputParser,
    ParserGenerator,
    AttributeDescription,
)
from debputy.manifest_parser.parser_doc import (
    render_rule,
    render_attribute_doc,
    doc_args_for_parser_doc,
)
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
    DeclarativeInputParser,
    DispatchingParserBase,
    InPackageContextParser,
    ListWrappedDeclarativeInputParser,
)
from debputy.util import _info, _warn, detect_possible_typo
from debputy.yaml import MANIFEST_YAML
from debputy.yaml.compat import (
    Node,
    CommentedMap,
    LineCol,
    CommentedSeq,
    CommentedBase,
    YAMLError,
)

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

try:
    from pygls.server import LanguageServer
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


YAML_COMPLETION_HINT_KEY = "___COMPLETE:"
YAML_COMPLETION_HINT_VALUE = "___COMPLETE"
DEBPUTY_PLUGIN_METADATA = plugin_metadata_for_debputys_own_plugin()


def resolve_hover_text_for_value(
    feature_set: PluginProvidedFeatureSet,
    parser: DeclarativeMappingInputParser,
    plugin_metadata: DebputyPluginMetadata,
    segment: Union[str, int],
    matched: Any,
) -> Optional[str]:

    hover_doc_text: Optional[str] = None
    attr = parser.manifest_attributes.get(segment)
    attr_type = attr.attribute_type if attr is not None else None
    if attr_type is None:
        _info(f"Matched value for {segment} -- No attr or type")
        return None
    if isinstance(attr_type, type) and issubclass(attr_type, DebputyDispatchableType):
        parser_generator = feature_set.manifest_parser_generator
        parser = parser_generator.dispatch_parser_table_for(attr_type)
        if parser is None or not isinstance(matched, str):
            _info(
                f"Unknown parser for {segment} or matched is not a str -- {attr_type} {type(matched)=}"
            )
            return None
        subparser = parser.parser_for(matched)
        if subparser is None:
            _info(f"Unknown parser for {matched} (subparser)")
            return None
        hover_doc_text = render_rule(
            matched,
            subparser.parser,
            plugin_metadata,
        )
    else:
        _info(f"Unknown value: {matched} -- {segment}")
    return hover_doc_text


def resolve_hover_text(
    feature_set: PluginProvidedFeatureSet,
    parser: Optional[Union[DeclarativeInputParser[Any], DispatchingParserBase]],
    plugin_metadata: DebputyPluginMetadata,
    segments: List[Union[str, int]],
    at_depth_idx: int,
    matched: Any,
    matched_key: bool,
) -> Optional[str]:
    hover_doc_text: Optional[str] = None
    if at_depth_idx == len(segments):
        segment = segments[at_depth_idx - 1]
        _info(f"Matched {segment} at ==, {matched_key=} ")
        hover_doc_text = render_rule(
            segment,
            parser,
            plugin_metadata,
            is_root_rule=False,
        )
    elif at_depth_idx + 1 == len(segments) and isinstance(
        parser, DeclarativeMappingInputParser
    ):
        segment = segments[at_depth_idx]
        _info(f"Matched {segment} at -1, {matched_key=} ")
        if isinstance(segment, str):
            if not matched_key:
                hover_doc_text = resolve_hover_text_for_value(
                    feature_set,
                    parser,
                    plugin_metadata,
                    segment,
                    matched,
                )
            if matched_key or hover_doc_text is None:
                rule_name = _guess_rule_name(segments, at_depth_idx)
                hover_doc_text = _render_param_doc(
                    rule_name,
                    parser,
                    plugin_metadata,
                    segment,
                )
    else:
        _info(f"No doc: {at_depth_idx=} {len(segments)=}")

    return hover_doc_text


def as_hover_doc(
    ls: "DebputyLanguageServer",
    hover_doc_text: Optional[str],
) -> Optional[types.Hover]:
    if hover_doc_text is None:
        return None
    return types.Hover(
        contents=types.MarkupContent(
            kind=ls.hover_markup_format(
                types.MarkupKind.Markdown,
                types.MarkupKind.PlainText,
            ),
            value=hover_doc_text,
        ),
    )


def _render_param_doc(
    rule_name: str,
    declarative_parser: DeclarativeMappingInputParser,
    plugin_metadata: DebputyPluginMetadata,
    attribute: str,
) -> Optional[str]:
    attr = declarative_parser.source_attributes.get(attribute)
    if attr is None:
        return None

    doc_args, parser_doc = doc_args_for_parser_doc(
        rule_name,
        declarative_parser,
        plugin_metadata,
    )
    rendered_docs = render_attribute_doc(
        declarative_parser,
        declarative_parser.source_attributes,
        declarative_parser.input_time_required_parameters,
        declarative_parser.at_least_one_of,
        parser_doc,
        doc_args,
        is_interactive=True,
        rule_name=rule_name,
    )

    for attributes, rendered_doc in rendered_docs:
        if attribute in attributes:
            full_doc = [
                f"# Attribute `{attribute}`",
                "",
            ]
            full_doc.extend(rendered_doc)

            return "\n".join(full_doc)
    return None


def _guess_rule_name(segments: List[Union[str, int]], idx: int) -> str:
    orig_idx = idx
    idx -= 1
    while idx >= 0:
        segment = segments[idx]
        if isinstance(segment, str):
            return segment
        idx -= 1
    _warn(f"Unable to derive rule name from {segments} [{orig_idx}]")
    return "<Bug: unknown rule name>"


def is_at(position: types.Position, lc_pos: Tuple[int, int]) -> bool:
    return position.line == lc_pos[0] and position.character == lc_pos[1]


def is_before(position: types.Position, lc_pos: Tuple[int, int]) -> bool:
    line, column = lc_pos
    if position.line < line:
        return True
    if position.line == line and position.character < column:
        return True
    return False


def is_after(position: types.Position, lc_pos: Tuple[int, int]) -> bool:
    line, column = lc_pos
    if position.line > line:
        return True
    if position.line == line and position.character > column:
        return True
    return False


def error_range_at_position(
    lines: List[str],
    line_no: int,
    char_offset: int,
) -> TERange:
    line = lines[line_no]
    line_len = len(line)
    start_idx = char_offset
    end_idx = start_idx

    if line[start_idx].isspace():

        def _check(x: str) -> bool:
            return not x.isspace()

    else:

        def _check(x: str) -> bool:
            return x.isspace()

    for i in range(end_idx, line_len):
        end_idx = i
        if _check(line[i]):
            break

    for i in range(start_idx, -1, -1):
        if i > 0 and _check(line[i]):
            break
        start_idx = i

    return TERange(
        TEPosition(line_no, start_idx),
        TEPosition(line_no, end_idx),
    )


def _escape(v: str) -> str:
    return '"' + v.replace("\n", "\\n") + '"'


def insert_complete_marker_snippet(
    lines: List[str],
    server_position: types.Position,
) -> bool:
    _info(f"Complete at {server_position}")
    line_no = server_position.line
    line = lines[line_no] if line_no < len(lines) else ""
    pos_rhs = (
        line[server_position.character :]
        if server_position.character < len(line)
        else ""
    )
    if pos_rhs and not pos_rhs.isspace():
        _info(f"No insertion: {_escape(line[server_position.character:])}")
        return False
    lhs_ws = line[: server_position.character]
    lhs = lhs_ws.strip()
    if lhs.endswith(":"):
        _info("Insertion of value (key seen)")
        new_line = line[: server_position.character] + YAML_COMPLETION_HINT_VALUE + "\n"
    elif lhs.startswith("-"):
        _info("Insertion of key or value (list item)")
        # Respect the provided indentation
        snippet = (
            YAML_COMPLETION_HINT_KEY if ":" not in lhs else YAML_COMPLETION_HINT_VALUE
        )
        new_line = line[: server_position.character] + snippet + "\n"
    elif not lhs or (lhs_ws and not lhs_ws[0].isspace()):
        _info(f"Insertion of key or value: {_escape(line[server_position.character:])}")
        # Respect the provided indentation
        snippet = (
            YAML_COMPLETION_HINT_KEY if ":" not in lhs else YAML_COMPLETION_HINT_VALUE
        )
        new_line = line[: server_position.character] + snippet + "\n"
    elif lhs.isalpha() and ":" not in lhs:
        _info(f"Expanding value to a key: {_escape(line[server_position.character:])}")
        # Respect the provided indentation
        new_line = line[: server_position.character] + YAML_COMPLETION_HINT_KEY + "\n"
    else:
        c = (
            line[server_position.character]
            if server_position.character < len(line)
            else "(OOB)"
        )
        _info(f"Not touching line: {_escape(line)} -- {_escape(c)}")
        return False
    _info(f'Evaluating complete on synthetic line: "{new_line}"')
    if line_no < len(lines):
        lines[line_no] = new_line
    elif line_no == len(lines):
        lines.append(new_line)
    else:
        return False
    return True


def yaml_key_range(
    key: Optional[str],
    line: int,
    col: int,
) -> "TERange":
    key_len = len(key) if key else 1
    return TERange.between(
        TEPosition(line, col),
        TEPosition(line, col + key_len),
    )


def yaml_flag_unknown_key(
    lint_state: LintState,
    key: Optional[str],
    expected_keys: Iterable[str],
    line: int,
    col: int,
    *,
    message_format: str = 'Unknown or unsupported key "{key}".',
    unknown_keys_diagnostic_severity: Optional[LintSeverity] = "error",
) -> Optional[str]:
    key_range = yaml_key_range(key, line, col)

    candidates = detect_possible_typo(key, expected_keys) if key is not None else ()
    extra = ""
    corrected_key = None
    if candidates:
        extra = f' It looks like a typo of "{candidates[0]}".'
        # TODO: We should be able to tell that `install-doc` and `install-docs` are the same.
        #  That would enable this to work in more cases.
        corrected_key = candidates[0] if len(candidates) == 1 else None
        if unknown_keys_diagnostic_severity is None:
            message_format = f"Possible typo of {candidates[0]}."
            extra = ""
    elif unknown_keys_diagnostic_severity is None:
        return None

    if key is None:
        message_format = "Missing key"
    if unknown_keys_diagnostic_severity is not None:
        lint_state.emit_diagnostic(
            key_range,
            message_format.format(key=key) + extra,
            unknown_keys_diagnostic_severity,
            "debputy",
            quickfixes=[propose_correct_text_quick_fix(n) for n in candidates],
        )
    return corrected_key


def resolve_keyword(
    current_parser: Union[DeclarativeInputParser[Any], DispatchingParserBase],
    current_plugin: DebputyPluginMetadata,
    segments: List[Union[str, int]],
    segment_idx: int,
    parser_generator: ParserGenerator,
    *,
    is_completion_attempt: bool = False,
) -> Optional[
    Tuple[
        Union[DeclarativeInputParser[Any], DispatchingParserBase],
        DebputyPluginMetadata,
        int,
    ]
]:
    if segment_idx >= len(segments):
        return current_parser, current_plugin, segment_idx
    current_segment = segments[segment_idx]
    if isinstance(current_parser, ListWrappedDeclarativeInputParser):
        if isinstance(current_segment, int):
            current_parser = current_parser.delegate
            segment_idx += 1
            if segment_idx >= len(segments):
                return current_parser, current_plugin, segment_idx
            current_segment = segments[segment_idx]

    if not isinstance(current_segment, str):
        return None

    if is_completion_attempt and current_segment.endswith(
        (YAML_COMPLETION_HINT_KEY, YAML_COMPLETION_HINT_VALUE)
    ):
        return current_parser, current_plugin, segment_idx

    if isinstance(current_parser, InPackageContextParser):
        return resolve_keyword(
            current_parser.delegate,
            current_plugin,
            segments,
            segment_idx + 1,
            parser_generator,
            is_completion_attempt=is_completion_attempt,
        )
    elif isinstance(current_parser, DispatchingParserBase):
        if not current_parser.is_known_keyword(current_segment):
            if is_completion_attempt:
                return current_parser, current_plugin, segment_idx
            return None
        subparser = current_parser.parser_for(current_segment)
        segment_idx += 1
        if segment_idx < len(segments):
            return resolve_keyword(
                subparser.parser,
                subparser.plugin_metadata,
                segments,
                segment_idx,
                parser_generator,
                is_completion_attempt=is_completion_attempt,
            )
        return subparser.parser, subparser.plugin_metadata, segment_idx
    elif isinstance(current_parser, DeclarativeMappingInputParser):
        attr = current_parser.manifest_attributes.get(current_segment)
        attr_type = attr.attribute_type if attr is not None else None
        if (
            attr_type is not None
            and isinstance(attr_type, type)
            and issubclass(attr_type, DebputyDispatchableType)
        ):
            subparser = parser_generator.dispatch_parser_table_for(attr_type)
            if subparser is not None and (
                is_completion_attempt or segment_idx + 1 < len(segments)
            ):
                return resolve_keyword(
                    subparser,
                    current_plugin,
                    segments,
                    segment_idx + 1,
                    parser_generator,
                    is_completion_attempt=is_completion_attempt,
                )
        return current_parser, current_plugin, segment_idx
    else:
        _info(f"Unknown parser: {current_parser.__class__}")
    return None


def _trace_cursor(
    content: Any,
    attribute_path: AttributePath,
    server_position: types.Position,
) -> Optional[Tuple[bool, AttributePath, Any, Any]]:
    matched_key: Optional[Union[str, int]] = None
    matched: Optional[Node] = None
    matched_was_key: bool = False

    if isinstance(content, CommentedMap):
        dict_lc: LineCol = content.lc
        for k, v in content.items():
            k_lc = dict_lc.key(k)
            if is_before(server_position, k_lc):
                break
            v_lc = dict_lc.value(k)
            if is_before(server_position, v_lc):
                # TODO: Handle ":" and "whitespace"
                matched = k
                matched_key = k
                matched_was_key = True
                break
            matched = v
            matched_key = k
    elif isinstance(content, CommentedSeq):
        list_lc: LineCol = content.lc
        for idx, value in enumerate(content):
            i_lc = list_lc.item(idx)
            if is_before(server_position, i_lc):
                break
            matched_key = idx
            matched = value

    if matched is not None:
        assert matched_key is not None
        sub_path = attribute_path[matched_key]
        if not matched_was_key and isinstance(matched, CommentedBase):
            return _trace_cursor(matched, sub_path, server_position)
        return matched_was_key, sub_path, matched, content
    return None


def maybe_quote_yaml_value(v: str) -> str:
    if v and v[0].isdigit():
        try:
            float(v)
            return f"'{v}'"
        except ValueError:
            pass
    return v


def _complete_value(v: Any) -> str:
    if isinstance(v, str):
        return maybe_quote_yaml_value(v)
    return str(v)


def completion_from_attr(
    attr: AttributeDescription,
    pg: ParserGenerator,
    matched: Any,
) -> Optional[Union[types.CompletionList, Sequence[types.CompletionItem]]]:
    type_mapping = pg.get_mapped_type_from_target_type(attr.attribute_type)
    if type_mapping is not None:
        attr_type = type_mapping.source_type
    else:
        attr_type = attr.attribute_type

    orig = get_origin(attr_type)
    valid_values: Sequence[Any] = tuple()

    if orig == Literal:
        valid_values = get_args(attr_type)
    elif orig == bool or attr.attribute_type == bool:
        valid_values = ("true", "false")
    elif isinstance(orig, type) and issubclass(orig, DebputyDispatchableType):
        parser = pg.dispatch_parser_table_for(orig)
        _info(f"M: {parser}")

    if matched in valid_values:
        _info(f"Already filled: {matched} is one of {valid_values}")
        return None
    if valid_values:
        return [types.CompletionItem(_complete_value(x)) for x in valid_values]
    return None


def generic_yaml_hover(
    ls: "DebputyLanguageServer",
    params: types.HoverParams,
    root_parser_initializer: Callable[
        [ParserGenerator], Union[DeclarativeInputParser[Any], DispatchingParserBase]
    ],
) -> Optional[types.Hover]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    position_codec = doc.position_codec
    server_position = position_codec.position_from_client_units(lines, params.position)

    try:
        content = MANIFEST_YAML.load("".join(lines))
    except YAMLError:
        return None
    attribute_root_path = AttributePath.root_path(content)
    m = _trace_cursor(content, attribute_root_path, server_position)
    if m is None:
        _info("No match")
        return None
    matched_key, attr_path, matched, _ = m
    _info(f"Matched path: {matched} (path: {attr_path.path}) [{matched_key=}]")

    feature_set = ls.plugin_feature_set
    parser_generator = feature_set.manifest_parser_generator
    root_parser = root_parser_initializer(parser_generator)
    segments = list(attr_path.path_segments())
    km = resolve_keyword(
        root_parser,
        DEBPUTY_PLUGIN_METADATA,
        segments,
        0,
        parser_generator,
    )
    if km is None:
        _info("No keyword match")
        return None
    parser, plugin_metadata, at_depth_idx = km
    _info(f"Match leaf parser {at_depth_idx}/{len(segments)} -- {parser.__class__}")
    hover_doc_text = resolve_hover_text(
        feature_set,
        parser,
        plugin_metadata,
        segments,
        at_depth_idx,
        matched,
        matched_key,
    )
    return as_hover_doc(ls, hover_doc_text)



"""Wrapper module around `lsprotocol.types`

This wrapper module is here to facility making `lsprotocol` optional
for -backports. When available, it is a (mostly) transparent wrapper
for the real module. When missing, it returns a placeholder object
for anything for the purpose of making things simpler.
"""

from typing import TYPE_CHECKING, Any, Iterator, List

if TYPE_CHECKING:
    from lsprotocol import types

    # To defeat "unused" detections that might attempt to
    # optimize out the import
    assert types is not None
    __all__ = dir(types)

else:
    try:
        from lsprotocol import types

        __all__ = dir(types)

    except ImportError:

        stub_attr = {
            "__name__": __name__,
            "__file__": __file__,
            "__doc__": __doc__,
        }
        bad_attr = frozenset(
            [
                "pytestmark",
                "pytest_plugins",
            ]
        )

        class StubModule:
            @staticmethod
            def __getattr__(item: Any) -> Any:
                if item in stub_attr:
                    return stub_attr[item]
                if item in bad_attr:
                    raise AttributeError(item)
                if isinstance(item, str) and item.startswith("TEXT_"):
                    return str(item)
                return types

            def __call__(self, *args, **kwargs) -> Any:
                return self

            def __iter__(self) -> Iterator[Any]:
                return iter(())

        types = StubModule()
        __all__ = []


def __dir__() -> List[str]:
    return dir(types)


def __getattr__(name: str) -> Any:
    return getattr(types, name)



import textwrap
from typing import cast, Any, TYPE_CHECKING

from debputy.plugin.api import (
    DebputyPluginInitializer,
    VirtualPath,
    BinaryCtrlAccessor,
    PackageProcessingContext,
)

from debputy.util import POSTINST_DEFAULT_CONDITION

if TYPE_CHECKING:
    from debputy.filesystem_scan import VirtualPathBase
    from debputy.plugin.api.impl import DebputyPluginInitializerProvider

PRIVATE_PYTHON_DIR = "/usr/share/debputy"


def _maintscript_generator(
    _path: VirtualPath,
    ctrl: BinaryCtrlAccessor,
    context: PackageProcessingContext,
) -> None:
    maintscript = ctrl.maintscript
    package_name = context.binary_package.name

    # When `debputy` becomes a stand-alone package, it should have these maintscripts instead of dh-debputy
    # Admittedly, I hope to get rid of this plugin before then, but ...
    assert package_name == "dh-debputy", "Update the self-hosting plugin"

    ctrl.dpkg_trigger("interest-noawait", PRIVATE_PYTHON_DIR)
    maintscript.unconditionally_in_script(
        "postinst",
        textwrap.dedent(
            f"""\
        if {POSTINST_DEFAULT_CONDITION} || [ "$1" = "triggered" ] ; then
            # Ensure all plugins are byte-compiled (plus uninstalled plugins are cleaned up)
            py3clean {PRIVATE_PYTHON_DIR}
            if command -v py3compile >/dev/null 2>&1; then
                py3compile {PRIVATE_PYTHON_DIR}
            fi
            if command -v pypy3compile >/dev/null 2>&1; then
                pypy3compile {PRIVATE_PYTHON_DIR} || true
            fi
        fi
    """
        ),
    )
    maintscript.unconditionally_in_script(
        "prerm",
        textwrap.dedent(
            f"""\
    if command -v py3clean >/dev/null 2>&1; then
        py3clean {PRIVATE_PYTHON_DIR}
    else
        find {PRIVATE_PYTHON_DIR}/ -type d -name __pycache__ -empty -print0 | xargs --null --no-run-if-empty rmdir
    fi
    """
        ),
    )


def _rtupdate_generator(
    fs_root: VirtualPath,
    _: Any,
    context: "PackageProcessingContext",
) -> None:
    package_name = context.binary_package.name
    # When `debputy` becomes a stand-alone package, it should have these scripts instead of dh-debputy
    # Admittedly, I hope to get rid of this plugin before then, but ...
    assert package_name == "dh-debputy", "Update the self-hosting plugin"
    rtupdate_dir: VirtualPathBase = cast(
        "VirtualPathBase",
        fs_root.mkdirs("/usr/share/python3/runtime.d/"),
    )
    with rtupdate_dir.open_child("debputy.rtupdate", "w") as fd:
        template = textwrap.dedent(
            f"""\
        #! /bin/sh
        if [ "$1" = rtupdate ]; then
            py3clean {PRIVATE_PYTHON_DIR}
            py3compile {PRIVATE_PYTHON_DIR}
        fi
        """
        )
        fd.write(template)
    rtupdate_dir["debputy.rtupdate"].chmod(0o755)


def initializer(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector(
        "debputy-self-hosting",
        _maintscript_generator,
    )
    internal_api: DebputyPluginInitializerProvider = cast(
        "DebputyPluginInitializerProvider", api
    )
    internal_api.package_processor(
        "rtupdate",
        _rtupdate_generator,
    )



import re
from typing import Any

from debputy.plugin.api import (
    DebputyPluginInitializer,
    BinaryCtrlAccessor,
    PackageProcessingContext,
)
from debputy.util import _error

GNOME_VERSION1_RE = re.compile(r"^(\d+:)?(\d+)\.(\d+)\.[\d.]+.*$")
GNOME_VERSION2_RE = re.compile(
    r"^(\d+:)?(\d+)(?:\.[\d.]+|~(alpha|beta|rc)[\d.]*|[+~])?.*$"
)


def initialize(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector(
        "gnome-versions",
        gnome_versions,
        # Probably not necessary, but this is the most faithful conversion
        package_type=["deb", "udeb"],
    )
    # Looking for "clean_la_files"? The `debputy` plugin provides a replacement
    # feature.


def gnome_versions(
    _unused: Any,
    ctrl: BinaryCtrlAccessor,
    context: PackageProcessingContext,
) -> None:
    # Conversion note: In debhelper, $dh{VERSION} is actually the "source" version
    # (though sometimes it has a binNMU version too).  In `debputy`, we have access
    # to the "true" binary version (dpkg-gencontrol -v<VERSION>). In 99% of all cases,
    # the difference is irrelevant as people rarely use dpkg-gencontrol -v<VERSION>.
    version = context.binary_package_version
    m = GNOME_VERSION1_RE.match(version)
    epoch = ""
    gnome_version = ""
    gnome_next_version = ""
    if m:
        major_version = int(m.group(2))
        if major_version < 40:
            epoch = m.group(1)
            minor_version = int(m.group(3))
            gnome_version = f"{major_version}.{minor_version}"
            if major_version == 3 and minor_version == 38:
                prefix = ""
            else:
                prefix = f"{major_version}."
            gnome_next_version = f"{prefix}{minor_version + 2}"
    if gnome_version == "":
        m = GNOME_VERSION2_RE.match(version)
        if not m:
            _error(
                f"Unable to determine the GNOME major version from {version} for package"
                f" {context.binary_package.name}. If this is not a GNOME package or it does"
                f" not follow the GNOME version standard, please disable the GNOME plugin"
                f" (debputy-plugin-gnome)."
            )
        epoch = m.group(1)
        version = int(m.group(2))
        gnome_version = f"{version}~"
        gnome_next_version = f"{version + 1}~"
    if epoch is None:
        epoch = ""
    ctrl.substvars["gnome:Version"] = f"{epoch}{gnome_version}"
    ctrl.substvars["gnome:UpstreamVersion"] = f"{gnome_version}"
    ctrl.substvars["gnome:NextVersion"] = f"{epoch}{gnome_next_version}"
    ctrl.substvars["gnome:NextUpstreamVersion"] = f"{gnome_next_version}"



import textwrap
from typing import List, Dict, Tuple, Mapping

from debian.deb822 import Deb822

from debputy.maintscript_snippet import MaintscriptSnippetContainer, MaintscriptSnippet
from debputy.packager_provided_files import PackagerProvidedFile
from debputy.packages import BinaryPackage
from debputy.packaging.makeshlibs import resolve_reserved_provided_file
from debputy.plugin.api import VirtualPath
from debputy.util import _error, escape_shell, POSTINST_DEFAULT_CONDITION

# Match debhelper (minus one space in each end, which comes
# via join).
LINE_PREFIX = "\\\n            "


def process_alternatives(
    binary_package: BinaryPackage,
    fs_root: VirtualPath,
    reserved_packager_provided_files: Dict[str, List[PackagerProvidedFile]],
    maintscript_snippets: Dict[str, MaintscriptSnippetContainer],
) -> None:
    if binary_package.is_udeb:
        return

    provided_alternatives_file = resolve_reserved_provided_file(
        "alternatives",
        reserved_packager_provided_files,
    )
    if provided_alternatives_file is None:
        return

    with provided_alternatives_file.open() as fd:
        alternatives = list(Deb822.iter_paragraphs(fd))

    for no, alternative in enumerate(alternatives):
        process_alternative(
            provided_alternatives_file.fs_path,
            fs_root,
            alternative,
            no,
            maintscript_snippets,
        )


def process_alternative(
    provided_alternatives_fs_path: str,
    fs_root: VirtualPath,
    alternative_deb822: Deb822,
    no: int,
    maintscript_snippets: Dict[str, MaintscriptSnippetContainer],
) -> None:
    name = _mandatory_key(
        "Name",
        alternative_deb822,
        provided_alternatives_fs_path,
        f"Stanza number {no}",
    )
    error_context = f"Alternative named {name}"
    link_path = _mandatory_key(
        "Link",
        alternative_deb822,
        provided_alternatives_fs_path,
        error_context,
    )
    impl_path = _mandatory_key(
        "Alternative",
        alternative_deb822,
        provided_alternatives_fs_path,
        error_context,
    )
    priority = _mandatory_key(
        "Priority",
        alternative_deb822,
        provided_alternatives_fs_path,
        error_context,
    )
    if "/" in name:
        _error(
            f'The "Name" ({link_path}) key must be a basename and cannot contain slashes'
            f" ({error_context} in {provided_alternatives_fs_path})"
        )
    if link_path == impl_path:
        _error(
            f'The "Link" key and the "Alternative" key must not have the same value'
            f" ({error_context} in {provided_alternatives_fs_path})"
        )
    impl = fs_root.lookup(impl_path)
    if impl is None or impl.is_dir:
        _error(
            f'The path listed in "Alternative" ("{impl_path}") does not exist'
            f" in the package. ({error_context} in {provided_alternatives_fs_path})"
        )
    for key in ["Slave", "Slaves", "Slave-Links"]:
        if key in alternative_deb822:
            _error(
                f'Please use "Dependents" instead of "{key}".'
                f" ({error_context} in {provided_alternatives_fs_path})"
            )
    dependents = alternative_deb822.get("Dependents")
    install_command = [
        escape_shell(
            "update-alternatives",
            "--install",
            link_path,
            name,
            impl_path,
            priority,
        )
    ]
    remove_command = [
        escape_shell(
            "update-alternatives",
            "--remove",
            name,
            impl_path,
        )
    ]
    if dependents:
        seen_link_path = set()
        for line in dependents.splitlines():
            line = line.strip()
            if not line:  # First line is usually empty
                continue
            dlink_path, dlink_name, dimpl_path = parse_dependent_link(
                line,
                error_context,
                provided_alternatives_fs_path,
            )
            if dlink_path in seen_link_path:
                _error(
                    f'The Dependent link path "{dlink_path}" was used twice.'
                    f" ({error_context} in {provided_alternatives_fs_path})"
                )
            dimpl = fs_root.lookup(dimpl_path)
            if dimpl is None or dimpl.is_dir:
                _error(
                    f'The path listed in "Dependents" ("{dimpl_path}") does not exist'
                    f" in the package. ({error_context} in {provided_alternatives_fs_path})"
                )
            seen_link_path.add(dlink_path)
            install_command.append(LINE_PREFIX)
            install_command.append(
                escape_shell(
                    # update-alternatives still uses this old option name :-/
                    "--slave",
                    dlink_path,
                    dlink_name,
                    dimpl_path,
                )
            )
    postinst = textwrap.dedent(
        """\
    if {CONDITION}; then
        {COMMAND}
    fi
    """
    ).format(
        CONDITION=POSTINST_DEFAULT_CONDITION,
        COMMAND=" ".join(install_command),
    )

    prerm = textwrap.dedent(
        """\
    if [ "$1" = "remove" ]; then
        {COMMAND}
    fi
    """
    ).format(COMMAND=" ".join(remove_command))
    maintscript_snippets["postinst"].append(
        MaintscriptSnippet(
            f"debputy (via {provided_alternatives_fs_path})",
            snippet=postinst,
        )
    )
    maintscript_snippets["prerm"].append(
        MaintscriptSnippet(
            f"debputy (via {provided_alternatives_fs_path})",
            snippet=prerm,
        )
    )


def parse_dependent_link(
    line: str,
    error_context: str,
    provided_alternatives_file: str,
) -> Tuple[str, str, str]:
    parts = line.split()
    if len(parts) != 3:
        if len(parts) > 1:
            pass
        _error(
            f"The each line in Dependents links must have exactly 3 space separated parts."
            f' The "{line}" split into {len(parts)} part(s).'
            f" ({error_context} in {provided_alternatives_file})"
        )

    dlink_path, dlink_name, dimpl_path = parts
    if "/" in dlink_name:
        _error(
            f'The Dependent link name "{dlink_path}" must be a basename and cannot contain slashes'
            f" ({error_context} in {provided_alternatives_file})"
        )
    if dlink_path == dimpl_path:
        _error(
            f'The Dependent Link path and Alternative must not have the same value ["{dlink_path}"]'
            f" ({error_context} in {provided_alternatives_file})"
        )
    return dlink_path, dlink_name, dimpl_path


def _mandatory_key(
    key: str,
    alternative_deb822: Mapping[str, str],
    provided_alternatives_file: str,
    error_context: str,
) -> str:
    try:
        return alternative_deb822[key]
    except KeyError:
        _error(
            f'Missing mandatory key "{key}" in {provided_alternatives_file} ({error_context})'
        )



import atexit
import gettext
import os
from functools import lru_cache
from gettext import NullTranslations, GNUTranslations
from tempfile import TemporaryDirectory
from typing import Optional, Union, Set, Iterable, TYPE_CHECKING

from debputy import DEBPUTY_IS_RUN_FROM_SOURCE, DEBPUTY_ROOT_DIR
from debputy.util import _debug_log

try:
    import polib

    HAS_POLIB = True
except ImportError:
    HAS_POLIB = False

if TYPE_CHECKING:
    import polib

Translations = Union[NullTranslations, GNUTranslations]


def N_(v: str) -> str:
    return v


@lru_cache
def _temp_messages_dir() -> str:
    temp_dir = TemporaryDirectory(delete=False, ignore_cleanup_errors=True)
    atexit.register(lambda: temp_dir.cleanup())
    return temp_dir.name


_GENERATED_MO_FILES_FOR: Set[str] = set()


def translation(
    domain: str,
    *,
    languages: Optional[Iterable[str]] = None,
) -> Translations:
    if DEBPUTY_IS_RUN_FROM_SOURCE and HAS_POLIB:
        po_domain_dir = DEBPUTY_ROOT_DIR / "po" / domain
        locale_dir = _temp_messages_dir()
        if po_domain_dir.is_dir() and domain not in _GENERATED_MO_FILES_FOR:
            for child in po_domain_dir.iterdir():
                if not child.is_file() or not child.name.endswith(".po"):
                    continue
                language = child.name[:-3]
                mo_dir = os.path.join(locale_dir, language, "LC_MESSAGES")
                os.makedirs(mo_dir, exist_ok=True)
                mo_path = os.path.join(mo_dir, f"{domain}.mo")
                parsed_po_file = polib.pofile(child)
                parsed_po_file.save_as_mofile(mo_path)
            _GENERATED_MO_FILES_FOR.add(domain)
        try:
            r = gettext.translation(
                domain,
                localedir=locale_dir,
                languages=languages,
                fallback=True,
            )
            _debug_log(f"Found in-source translation for {domain}")
            return r
        except FileNotFoundError:
            # Fall through
            _debug_log(
                f"No in-source translation for {domain}, trying to installed translations"
            )

    # For some reason, the type checking thinks that `fallback` must be `Literal[False]` which
    # defeats the purpose of being able to provide a different value than the default for it.
    #
    # So `type: ignore` to "fix" that for now.
    return gettext.translation(
        domain,
        languages=languages,
        fallback=True,
    )


__all__ = ["N_", "translation", "Translations"]



import functools
import importlib.resources
import itertools
import os
import re
import subprocess
from importlib.resources.abc import Traversable
from typing import Iterable, FrozenSet, Tuple, Optional, Container, TYPE_CHECKING

from debian.debian_support import Release
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debputy.util import _info, _warn
import debputy.lsp.data.wordlists as data_wordlist

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )
except ImportError:
    pass

if TYPE_CHECKING:
    from debputy.linting.lint_util import LintState


_SPELL_CHECKER_DICT = "/usr/share/hunspell/en_US.dic"
_SPELL_CHECKER_AFF = "/usr/share/hunspell/en_US.aff"
_WORD_PARTS = re.compile(r"(\S+)")
_PRUNE_SYMBOLS_RE = re.compile(r"(\w+(?:-\w+|'\w+)?)")
_FIND_QUOTE_CHAR = re.compile(r'["`]')
_LOOKS_LIKE_FILENAME = re.compile(
    r"""
      [.]{0,3}/[a-z0-9]+(/[a-z0-9]+)+/*
    | [a-z0-9-_]+(/[a-z0-9]+)+/*
    | [a-z0-9_]+(/[a-z0-9_]+){2,}/*
    | (?:\S+)?[.][a-z]{1,3}

""",
    re.VERBOSE,
)
_LOOKS_LIKE_PROGRAMMING_TERM = re.compile(
    r"""
    (
        # Java identifier Camel Case
          [a-z][a-z0-9]*(?:[A-Z]{1,3}[a-z0-9]+)+
        # Type name Camel Case
        | [A-Z]{1,3}[a-z0-9]+(?:[A-Z]{1,3}[a-z0-9]+)+
        # Type name Camel Case with underscore (seen in Dh_Lib.pm among other
        | [A-Z]{1,3}[a-z0-9]+(?:_[A-Z]{1,3}[a-z0-9]+)+
        # Perl module
        | [A-Z]{1,3}[a-z0-9]+(?:_[A-Z]{1,3}[a-z0-9]+)*(::[A-Z]{1,3}[a-z0-9]+(?:_[A-Z]{1,3}[a-z0-9]+)*)+
        # Probably an abbreviation
        | [A-Z]{3,}
        # Perl/Python identifiers or Jinja templates
        | [$%&@_]?[{]?[{]?[a-z][a-z0-9]*(?:_[a-z0-9]+)+(?:(?:->)?[\[{]\S+|}}?)?
        # SCREAMING_SNAKE_CASE (environment variables plus -DVAR=B or $FOO)
        | [-$%&*_]{0,2}[A-Z][A-Z0-9]*(_[A-Z0-9]+)+(?:=\S+)?
        | \#[A-Z][A-Z0-9]*(_[A-Z0-9]+)+\#
        # Subcommand names. Require at least two "-" to avoid skipping hyphenated words
        | [a-z][a-z0-9]*(-[a-z0-9]+){2,}
        # Short args
        | -[a-z0-9]+
        # Things like 32bit
        | \d{2,}-?[a-z]+
        # Source package (we do not have a package without prefix/suffix because it covers 95% of all lowercase words)
        | src:[a-z0-9][-+.a-z0-9]+
        | [a-z0-9][-+.a-z0-9]+:(?:any|native)
        # Version
        | v\d+(?:[.]\S+)?
        # chmod symbolic mode or math
        | \S*=\S+
    )
""",
    re.VERBOSE,
)
_LOOKS_LIKE_EMAIL = re.compile(
    r"""
    <[^>@\s]+@[^>@\s]+>
""",
    re.VERBOSE,
)
_NO_CORRECTIONS = tuple()
_WORDLISTS = [
    "debian-wordlist.dic",
]
_NAMELISTS = [
    "logins-and-people.dic",
]
_PERSONAL_DICTS = [
    "${HOME}/.hunspell_default",
    "${HOME}/.hunspell_en_US",
]


try:
    if not os.path.lexists(_SPELL_CHECKER_DICT) or not os.path.lexists(
        _SPELL_CHECKER_AFF
    ):
        raise ImportError
    from hunspell import HunSpell

    _HAS_HUNSPELL = True
except ImportError:
    _HAS_HUNSPELL = False


def _read_wordlist(
    base_dir: Traversable,
    wordlist_name: str,
    *,
    namelist: bool = False,
) -> Iterable[str]:
    path = base_dir.joinpath(wordlist_name)
    with path.open("r", encoding="utf-8") as fd:
        w = [w.strip() for w in fd]
        yield from w
        if namelist:
            yield from (f"{n}'s" for n in w)


def _all_debian_archs() -> Iterable[str]:
    try:
        output = subprocess.check_output(["dpkg-architecture", "-L"])
    except (FileNotFoundError, subprocess.CalledProcessError) as e:
        _warn(f"dpkg-architecture -L failed: {e}")
        return tuple()

    return (x.strip() for x in output.decode("utf-8").splitlines())


@functools.lru_cache
def _builtin_exception_words() -> FrozenSet[str]:
    release_names = (x for x in Release.releases)
    data_dir = importlib.resources.files(data_wordlist.__name__)
    return frozenset(
        itertools.chain(
            itertools.chain.from_iterable(
                _read_wordlist(data_dir, wl) for wl in _WORDLISTS
            ),
            itertools.chain.from_iterable(
                _read_wordlist(data_dir, wl, namelist=True) for wl in _NAMELISTS
            ),
            release_names,
            _all_debian_archs(),
        )
    )


_DEFAULT_SPELL_CHECKER: Optional["Spellchecker"] = None


def spellcheck_line(
    lint_state: "LintState",
    line_no: int,
    line: str,
) -> int:
    spell_checker = lint_state.spellchecker()
    typos = 0
    for word, pos, endpos in spell_checker.iter_words(line):
        corrections = spell_checker.provide_corrections_for(word)
        if not corrections:
            continue
        word_range = TERange(
            TEPosition(line_no, pos),
            TEPosition(line_no, endpos),
        )
        typos += 1
        lint_state.emit_diagnostic(
            word_range,
            f'Spelling "{word}"',
            "spelling",
            "debputy",
            quickfixes=[propose_correct_text_quick_fix(c) for c in corrections],
            enable_non_interactive_auto_fix=False,
        )
    return typos


def default_spellchecker() -> "Spellchecker":
    global _DEFAULT_SPELL_CHECKER
    spellchecker = _DEFAULT_SPELL_CHECKER
    if spellchecker is None:
        if _HAS_HUNSPELL:
            spellchecker = HunspellSpellchecker()
        else:
            spellchecker = _do_nothing_spellchecker()
        _DEFAULT_SPELL_CHECKER = spellchecker
    return spellchecker


@functools.lru_cache()
def _do_nothing_spellchecker() -> "Spellchecker":
    return EverythingIsCorrectSpellchecker()


def disable_spellchecking() -> None:
    global _DEFAULT_SPELL_CHECKER
    _DEFAULT_SPELL_CHECKER = _do_nothing_spellchecker()


def _testing_set_default_spellchecker(
    spellchecker: Optional["Spellchecker"],
) -> Optional["Spellchecker"]:
    global _DEFAULT_SPELL_CHECKER
    previous = _DEFAULT_SPELL_CHECKER
    _DEFAULT_SPELL_CHECKER = spellchecker
    return previous


def _skip_quoted_parts(line: str) -> Iterable[Tuple[str, int]]:
    current_pos = 0
    while True:
        try:
            m = _FIND_QUOTE_CHAR.search(line, current_pos)
            if m is None:
                if current_pos == 0:
                    yield line, 0
                else:
                    yield line[current_pos:], current_pos
                return
            starting_marker_pos = m.span()[0]
            quote_char = m.group()
            end_marker_pos = line.index(quote_char, starting_marker_pos + 1)
        except ValueError:
            yield line[current_pos:], current_pos
            return

        part = line[current_pos:starting_marker_pos]

        if not part.isspace():
            yield part, current_pos
        current_pos = end_marker_pos + 1


def _split_line_to_words(line: str) -> Iterable[Tuple[str, int, int]]:
    for line_part, part_pos in _skip_quoted_parts(line):
        for m in _WORD_PARTS.finditer(line_part):
            fullword = m.group(1)
            if fullword.startswith("--"):
                # CLI arg
                continue
            if _LOOKS_LIKE_PROGRAMMING_TERM.match(fullword):
                continue
            if _LOOKS_LIKE_FILENAME.match(fullword):
                continue
            if _LOOKS_LIKE_EMAIL.match(fullword):
                continue
            mpos = m.span(1)[0]
            for sm in _PRUNE_SYMBOLS_RE.finditer(fullword):
                pos, endpos = sm.span(1)
                offset = part_pos + mpos
                yield sm.group(1), pos + offset, endpos + offset


class Spellchecker:

    @staticmethod
    def do_nothing_spellchecker() -> "Spellchecker":
        return EverythingIsCorrectSpellchecker()

    def iter_words(self, line: str) -> Iterable[Tuple[str, int, int]]:
        yield from _split_line_to_words(line)

    def provide_corrections_for(self, word: str) -> Iterable[str]:
        raise NotImplementedError

    def context_ignored_words(self, words: Container[str]) -> "Spellchecker":
        if not words:
            return self
        return ContextIgnoredWordsSpellchecker(self, words)


class ContextIgnoredWordsSpellchecker(Spellchecker):

    def __init__(
        self, spellchecker: Spellchecker, context_ignored_words: Container[str]
    ) -> None:
        self._spellchecker = spellchecker
        self._context_ignored_words = context_ignored_words

    def iter_words(self, line: str) -> Iterable[Tuple[str, int, int]]:
        return self._spellchecker.iter_words(line)

    def provide_corrections_for(self, word: str) -> Iterable[str]:
        if word.lower() in self._context_ignored_words:
            return _NO_CORRECTIONS
        return self._spellchecker.provide_corrections_for(word)


class EverythingIsCorrectSpellchecker(Spellchecker):
    def provide_corrections_for(self, word: str) -> Iterable[str]:
        return _NO_CORRECTIONS

    def context_ignored_words(self, words: Container[str]) -> "Spellchecker":
        return self


class HunspellSpellchecker(Spellchecker):

    def __init__(self) -> None:
        self._checker = HunSpell(_SPELL_CHECKER_DICT, _SPELL_CHECKER_AFF)
        for w in _builtin_exception_words():
            self._checker.add(w)
        self._load_personal_exclusions()

    def provide_corrections_for(self, word: str) -> Iterable[str]:
        if word.startswith(
            (
                "dpkg-",
                "dh-",
                "dh_",
                "debian-",
                "debconf-",
                "update-",
                "DEB_",
                "DPKG_",
            )
        ):
            return _NO_CORRECTIONS
        # 'ing is deliberately forcing a word into another word-class
        if word.endswith(("'ing", "-nss")):
            return _NO_CORRECTIONS
        return self._lookup(word)

    @functools.lru_cache(128)
    def _lookup(self, word: str) -> Iterable[str]:
        if self._checker.spell(word):
            return _NO_CORRECTIONS
        return self._checker.suggest(word)

    def _load_personal_exclusions(self) -> None:
        for filename in _PERSONAL_DICTS:
            if filename.startswith("${"):
                end_index = filename.index("}")
                varname = filename[2:end_index]
                value = os.environ.get(varname)
                if value is None:
                    continue
                filename = value + filename[end_index + 1 :]
            if os.path.isfile(filename):
                _info(f"Loading personal spelling dictionary from {filename}")
                self._checker.add_dic(filename)



#!/usr/bin/python3 -B
import argparse
import errno
import operator
import os
import stat
import subprocess
import tarfile
import textwrap
from typing import Optional, List, FrozenSet, Iterable, Callable, BinaryIO, cast

from debputy.intermediate_manifest import TarMember, PathType
from debputy.util import (
    _error,
    compute_output_filename,
    resolve_source_date_epoch,
    ColorizedArgumentParser,
    setup_logging,
    program_name,
    assume_not_none,
)
from debputy.version import __version__


# AR header / start of a deb file for reference
# 00000000  21 3c 61 72 63 68 3e 0a  64 65 62 69 61 6e 2d 62  |!<arch>.debian-b|
# 00000010  69 6e 61 72 79 20 20 20  31 36 36 38 39 37 33 36  |inary   16689736|
# 00000020  39 35 20 20 30 20 20 20  20 20 30 20 20 20 20 20  |95  0     0     |
# 00000030  31 30 30 36 34 34 20 20  34 20 20 20 20 20 20 20  |100644  4       |
# 00000040  20 20 60 0a 32 2e 30 0a  63 6f 6e 74 72 6f 6c 2e  |  `.2.0.control.|
# 00000050  74 61 72 2e 78 7a 20 20  31 36 36 38 39 37 33 36  |tar.xz  16689736|
# 00000060  39 35 20 20 30 20 20 20  20 20 30 20 20 20 20 20  |95  0     0     |
# 00000070  31 30 30 36 34 34 20 20  39 33 36 38 20 20 20 20  |100644  9368    |
# 00000080  20 20 60 0a fd 37 7a 58  5a 00 00 04 e6 d6 b4 46  |  `..7zXZ......F|


class ArMember:
    def __init__(
        self,
        name: str,
        mtime: int,
        fixed_binary: Optional[bytes] = None,
        write_to_impl: Optional[Callable[[BinaryIO], None]] = None,
    ) -> None:
        self.name = name
        self._mtime = mtime
        self._write_to_impl = write_to_impl
        self.fixed_binary = fixed_binary

    @property
    def is_fixed_binary(self) -> bool:
        return self.fixed_binary is not None

    @property
    def mtime(self) -> int:
        return self.mtime

    def write_to(self, fd: BinaryIO) -> None:
        writer = self._write_to_impl
        assert writer is not None
        writer(fd)


AR_HEADER_LEN = 60
AR_HEADER = b" " * AR_HEADER_LEN


def write_header(
    fd: BinaryIO,
    member: ArMember,
    member_len: int,
    mtime: int,
) -> None:
    header = b"%-16s%-12d0     0     100644  %-10d\x60\n" % (
        member.name.encode("ascii"),
        mtime,
        member_len,
    )
    fd.write(header)


def generate_ar_archive(
    output_filename: str,
    mtime: int,
    members: Iterable[ArMember],
    prefer_raw_exceptions: bool,
) -> None:
    try:
        with open(output_filename, "wb", buffering=0) as fd:
            fd.write(b"!<arch>\n")
            for member in members:
                if member.is_fixed_binary:
                    fixed_binary = assume_not_none(member.fixed_binary)
                    write_header(fd, member, len(fixed_binary), mtime)
                    fd.write(fixed_binary)
                else:
                    header_pos = fd.tell()
                    fd.write(AR_HEADER)
                    member.write_to(fd)
                    current_pos = fd.tell()
                    fd.seek(header_pos, os.SEEK_SET)
                    content_len = current_pos - header_pos - AR_HEADER_LEN
                    assert content_len >= 0
                    write_header(fd, member, content_len, mtime)
                    fd.seek(current_pos, os.SEEK_SET)
    except OSError as e:
        if prefer_raw_exceptions:
            raise
        if e.errno == errno.ENOSPC:
            _error(
                f"Unable to write {output_filename}.  The file system device reported disk full: {str(e)}"
            )
        elif e.errno == errno.EIO:
            _error(
                f"Unable to write {output_filename}.  The file system reported a generic I/O error: {str(e)}"
            )
        elif e.errno == errno.EROFS:
            _error(
                f"Unable to write {output_filename}.  The file system is read-only: {str(e)}"
            )
        raise
    print(f"Generated {output_filename}")


def _generate_tar_file(
    tar_members: Iterable[TarMember],
    compression_cmd: List[str],
    write_to: BinaryIO,
) -> None:
    with (
        subprocess.Popen(
            compression_cmd, stdin=subprocess.PIPE, stdout=write_to
        ) as compress_proc,
        tarfile.open(
            mode="w|",
            fileobj=compress_proc.stdin,
            format=tarfile.GNU_FORMAT,
            errorlevel=1,
        ) as tar_fd,
    ):
        for tar_member in tar_members:
            tar_info: tarfile.TarInfo = tar_member.create_tar_info(tar_fd)
            if tar_member.path_type == PathType.FILE:
                with open(assume_not_none(tar_member.fs_path), "rb") as mfd:
                    tar_fd.addfile(tar_info, fileobj=mfd)
            else:
                tar_fd.addfile(tar_info)
    compress_proc.wait()
    if compress_proc.returncode != 0:
        _error(
            f"Compression command {compression_cmd} failed with code {compress_proc.returncode}"
        )


def generate_tar_file_member(
    tar_members: Iterable[TarMember],
    compression_cmd: List[str],
) -> Callable[[BinaryIO], None]:
    def _impl(fd: BinaryIO) -> None:
        _generate_tar_file(
            tar_members,
            compression_cmd,
            fd,
        )

    return _impl


def _xz_cmdline(
    compression_rule: "Compression",
    parsed_args: Optional[argparse.Namespace],
) -> List[str]:
    compression_level = compression_rule.effective_compression_level(parsed_args)
    cmdline = ["xz", "-T2", "-" + str(compression_level)]
    strategy = None if parsed_args is None else parsed_args.compression_strategy
    if strategy is None:
        strategy = "none"
    if strategy != "none":
        cmdline.append("--" + strategy)
    cmdline.append("--no-adjust")
    return cmdline


def _gzip_cmdline(
    compression_rule: "Compression",
    parsed_args: Optional[argparse.Namespace],
) -> List[str]:
    compression_level = compression_rule.effective_compression_level(parsed_args)
    cmdline = ["gzip", "-n" + str(compression_level)]
    strategy = None if parsed_args is None else parsed_args.compression_strategy
    if strategy is not None and strategy != "none":
        raise ValueError(
            f"Not implemented: Compression strategy {strategy}"
            " for gzip is currently unsupported (but dpkg-deb does)"
        )
    return cmdline


def _uncompressed_cmdline(
    _unused_a: "Compression",
    _unused_b: Optional[argparse.Namespace],
) -> List[str]:
    return ["cat"]


class Compression:
    def __init__(
        self,
        default_compression_level: int,
        extension: str,
        allowed_strategies: FrozenSet[str],
        cmdline_builder: Callable[
            ["Compression", Optional[argparse.Namespace]], List[str]
        ],
    ) -> None:
        self.default_compression_level = default_compression_level
        self.extension = extension
        self.allowed_strategies = allowed_strategies
        self.cmdline_builder = cmdline_builder

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.extension}>"

    def effective_compression_level(
        self, parsed_args: Optional[argparse.Namespace]
    ) -> int:
        if parsed_args and parsed_args.compression_level is not None:
            return cast("int", parsed_args.compression_level)
        return self.default_compression_level

    def as_cmdline(self, parsed_args: Optional[argparse.Namespace]) -> List[str]:
        return self.cmdline_builder(self, parsed_args)

    def with_extension(self, filename: str) -> str:
        return filename + self.extension


COMPRESSIONS = {
    "xz": Compression(6, ".xz", frozenset({"none", "extreme"}), _xz_cmdline),
    "gzip": Compression(
        9,
        ".gz",
        frozenset({"none", "filtered", "huffman", "rle", "fixed"}),
        _gzip_cmdline,
    ),
    "none": Compression(0, "", frozenset({"none"}), _uncompressed_cmdline),
}


def _normalize_compression_args(parsed_args: argparse.Namespace) -> argparse.Namespace:
    if (
        parsed_args.compression_level == 0
        and parsed_args.compression_algorithm == "gzip"
    ):
        print(
            "Note: Mapping compression algorithm to none for compatibility with dpkg-deb (due to -Zgzip -z0)"
        )
        setattr(parsed_args, "compression_algorithm", "none")

    compression = COMPRESSIONS[parsed_args.compression_algorithm]
    strategy = parsed_args.compression_strategy
    if strategy is not None and strategy not in compression.allowed_strategies:
        _error(
            f'Compression algorithm "{parsed_args.compression_algorithm}" does not support compression strategy'
            f' "{strategy}".  Allowed values: {", ".join(sorted(compression.allowed_strategies))}'
        )
    return parsed_args


def parse_args() -> argparse.Namespace:
    try:
        compression_level_default = int(os.environ["DPKG_DEB_COMPRESSOR_LEVEL"])
    except (KeyError, ValueError):
        compression_level_default = None

    try:
        compression_type = os.environ["DPKG_DEB_COMPRESSOR_TYPE"]
    except (KeyError, ValueError):
        compression_type = "xz"

    try:
        threads_max = int(os.environ["DPKG_DEB_THREADS_MAX"])
    except (KeyError, ValueError):
        threads_max = None

    description = textwrap.dedent(
        """\
    THIS IS A PROTOTYPE "dpkg-deb -b" emulator with basic manifest support

    DO NOT USE THIS TOOL DIRECTLY.  It has not stability guarantees and will be removed as
    soon as "dpkg-deb -b" grows support for the relevant features.

    This tool is a prototype "dpkg-deb -b"-like interface for compiling a Debian package
    without requiring root even for static ownership.  It is a temporary stand-in for
    "dpkg-deb -b" until "dpkg-deb -b" will get support for a manifest.

    The tool operates on an internal JSON based manifest for now, because it was faster
    than building an mtree parser (which is the format that dpkg will likely end up
    using).

    As the tool is not meant to be used directly, it is full of annoying paper cuts that
    I refuse to fix or maintain. Use the high level tool instead.

    """
    )

    parser = ColorizedArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        allow_abbrev=False,
        prog=program_name(),
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "package_root_dir",
        metavar="PACKAGE_ROOT_DIR",
        help="Root directory of the package. Must contain a DEBIAN directory",
    )
    parser.add_argument(
        "package_output_path",
        metavar="PATH",
        help="Path where the package should be placed.  If it is directory,"
        " the base name will be determined from the package metadata",
    )

    parser.add_argument(
        "--intermediate-package-manifest",
        dest="package_manifest",
        metavar="JSON_FILE",
        action="store",
        default=None,
        help="INTERMEDIATE package manifest (JSON!)",
    )
    parser.add_argument(
        "--root-owner-group",
        dest="root_owner_group",
        action="store_true",
        help="Ignored. Accepted for compatibility with dpkg-deb -b",
    )
    parser.add_argument(
        "-b",
        "--build",
        dest="build_param",
        action="store_true",
        help="Ignored. Accepted for compatibility with dpkg-deb",
    )
    parser.add_argument(
        "--source-date-epoch",
        dest="source_date_epoch",
        action="store",
        type=int,
        default=None,
        help="Source date epoch (can also be given via the SOURCE_DATE_EPOCH environ variable",
    )
    parser.add_argument(
        "-Z",
        dest="compression_algorithm",
        choices=COMPRESSIONS,
        default=compression_type,
        help="The compression algorithm to be used",
    )
    parser.add_argument(
        "-z",
        dest="compression_level",
        metavar="{0-9}",
        choices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        default=compression_level_default,
        type=int,
        help="The compression level to be used",
    )
    parser.add_argument(
        "-S",
        dest="compression_strategy",
        # We have a different default for xz when strategy is unset and we are building a udeb
        action="store",
        default=None,
        help="The compression algorithm to be used. Concrete values depend on the compression"
        ' algorithm, but the value "none" is always allowed',
    )
    parser.add_argument(
        "--uniform-compression",
        dest="uniform_compression",
        action="store_true",
        default=True,
        help="Whether to use the same compression for the control.tar and the data.tar."
        " The default is to use uniform compression.",
    )
    parser.add_argument(
        "--no-uniform-compression",
        dest="uniform_compression",
        action="store_false",
        default=True,
        help="Disable uniform compression (see --uniform-compression)",
    )
    parser.add_argument(
        "--threads-max",
        dest="threads_max",
        default=threads_max,
        # TODO: Support this properly
        type=int,
        help="Ignored; accepted for compatibility",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug_mode",
        action="store_true",
        default=False,
        help="Enable debug logging and raw stack traces on errors",
    )

    parsed_args = parser.parse_args()
    parsed_args = _normalize_compression_args(parsed_args)

    return parsed_args


def _ctrl_member(
    member_path: str,
    fs_path: Optional[str] = None,
    path_type: PathType = PathType.FILE,
    mode: int = 0o644,
    mtime: int = 0,
) -> TarMember:
    if fs_path is None:
        assert member_path.startswith("./")
        fs_path = "DEBIAN" + member_path[1:]
    return TarMember(
        member_path=member_path,
        path_type=path_type,
        fs_path=fs_path,
        mode=mode,
        owner="root",
        uid=0,
        group="root",
        gid=0,
        mtime=mtime,
    )


CTRL_MEMBER_SCRIPTS = {
    "postinst",
    "preinst",
    "postrm",
    "prerm",
    "config",
    "isinstallable",
}


def _ctrl_tar_members(package_root_dir: str, mtime: int) -> Iterable[TarMember]:
    debian_root = os.path.join(package_root_dir, "DEBIAN")
    dir_st = os.stat(debian_root)
    dir_mtime = int(dir_st.st_mtime)
    yield _ctrl_member(
        "./",
        debian_root,
        path_type=PathType.DIRECTORY,
        mode=0o0755,
        mtime=min(mtime, dir_mtime),
    )
    with os.scandir(debian_root) as dir_iter:
        for ctrl_member in sorted(dir_iter, key=operator.attrgetter("name")):
            st = os.stat(ctrl_member)
            if not stat.S_ISREG(st.st_mode):
                _error(
                    f"{ctrl_member.path} is not a file and all control.tar members ought to be files!"
                )
            file_mtime = int(st.st_mtime)
            yield _ctrl_member(
                f"./{ctrl_member.name}",
                path_type=PathType.FILE,
                fs_path=ctrl_member.path,
                mode=0o0755 if ctrl_member.name in CTRL_MEMBER_SCRIPTS else 0o0644,
                mtime=min(mtime, file_mtime),
            )


def parse_manifest(manifest_path: "Optional[str]") -> "List[TarMember]":
    if manifest_path is None:
        _error(f"--intermediate-package-manifest is mandatory for now")
    return TarMember.parse_intermediate_manifest(manifest_path)


def main() -> None:
    setup_logging()
    parsed_args = parse_args()
    root_dir: str = parsed_args.package_root_dir
    output_path: str = parsed_args.package_output_path
    mtime = resolve_source_date_epoch(parsed_args.source_date_epoch)

    data_compression: Compression = COMPRESSIONS[parsed_args.compression_algorithm]
    data_compression_cmd = data_compression.as_cmdline(parsed_args)
    if parsed_args.uniform_compression:
        ctrl_compression = data_compression
        ctrl_compression_cmd = data_compression_cmd
    else:
        ctrl_compression = COMPRESSIONS["gzip"]
        ctrl_compression_cmd = COMPRESSIONS["gzip"].as_cmdline(None)

    if output_path.endswith("/") or os.path.isdir(output_path):
        deb_file = os.path.join(
            output_path,
            compute_output_filename(os.path.join(root_dir, "DEBIAN"), False),
        )
    else:
        deb_file = output_path

    pack(
        deb_file,
        ctrl_compression,
        data_compression,
        root_dir,
        parsed_args.package_manifest,
        mtime,
        ctrl_compression_cmd,
        data_compression_cmd,
        prefer_raw_exceptions=not parsed_args.debug_mode,
    )


def pack(
    deb_file: str,
    ctrl_compression: Compression,
    data_compression: Compression,
    root_dir: str,
    package_manifest: "Optional[str]",
    mtime: int,
    ctrl_compression_cmd: List[str],
    data_compression_cmd: List[str],
    prefer_raw_exceptions: bool = False,
) -> None:
    data_tar_members = parse_manifest(package_manifest)
    members = [
        ArMember("debian-binary", mtime, fixed_binary=b"2.0\n"),
        ArMember(
            ctrl_compression.with_extension("control.tar"),
            mtime,
            write_to_impl=generate_tar_file_member(
                _ctrl_tar_members(root_dir, mtime),
                ctrl_compression_cmd,
            ),
        ),
        ArMember(
            data_compression.with_extension("data.tar"),
            mtime,
            write_to_impl=generate_tar_file_member(
                data_tar_members,
                data_compression_cmd,
            ),
        ),
    ]
    generate_ar_archive(deb_file, mtime, members, prefer_raw_exceptions)


if __name__ == "__main__":
    main()



import textwrap
from typing import Type, Sequence, Mapping, Container, Iterable, Any

from debputy.manifest_parser.base_types import DebputyParsedContentStandardConditional
from debputy.manifest_parser.tagging_types import DebputyParsedContent
from debputy.plugin.api.spec import (
    ParserAttributeDocumentation,
    StandardParserAttributeDocumentation,
)
from debputy.plugin.debputy.to_be_api_types import (
    OptionalInstallDirectly,
    OptionalInSourceBuild,
    OptionalBuildDirectory,
    BuildRuleParsedFormat,
)

_STD_ATTR_DOCS: Mapping[
    Type[DebputyParsedContent],
    Sequence[ParserAttributeDocumentation],
] = {
    BuildRuleParsedFormat: [
        StandardParserAttributeDocumentation(
            frozenset(["name"]),
            textwrap.dedent(
                """\
          The name of the build step.

          The name is used for multiple things, such as:
            1) If you ever need to reference the build elsewhere, the name will be used.
            2) When `debputy` references the build in log output and error, it will use the name.
            3) It is used as defaults for when `debputy` derives build and `DESTDIR` directories
               for the build.
        """
            ),
            # Put in top,
            sort_category=-1000,
        ),
        StandardParserAttributeDocumentation(
            frozenset(["for_packages"]),
            textwrap.dedent(
                """\
          Which package or packages this build step applies to.

          Either a package name or a list of package names.
        """
            ),
        ),
        StandardParserAttributeDocumentation(
            frozenset(["environment"]),
            textwrap.dedent(
                """\
          Specify that this build step uses the named environment

          If omitted, the default environment will be used. If no default environment is present,
          then this option is mandatory.
        """
            ),
        ),
    ],
    OptionalBuildDirectory: [
        StandardParserAttributeDocumentation(
            frozenset(["build_directory"]),
            textwrap.dedent(
                """\
          The build directory to use for the build.

          By default, `debputy` will derive a build directory automatically if the build system needs
          it. However, it can be useful if you need to reference the directory name from other parts
          of the manifest or want a "better" name than `debputy` comes up with.
        """
            ),
        ),
    ],
    OptionalInSourceBuild: [
        StandardParserAttributeDocumentation(
            frozenset(["perform_in_source_build"]),
            textwrap.dedent(
                """\
          Whether the build system should use "in source" or "out of source" build.

          This is mostly useful for forcing "in source" builds for build systems that default to
          "out of source" builds like `autoconf`.

          The default depends on the build system and the value of the `build-directory` attribute
          (if supported by the build system).
        """
            ),
            # Late
            sort_category=500,
        ),
    ],
    OptionalInstallDirectly: [
        StandardParserAttributeDocumentation(
            frozenset(["install_directly_to_package"]),
            textwrap.dedent(
                """\
          Whether the build system should install all upstream content directly into the package.

          This option is mostly useful for disabling said behavior by setting the attribute to `false`.
          The attribute conditionally defaults to `true` when the build only applies to one package.
          If explicitly set to `true`, then this build step must apply to exactly one package (usually
          implying that `for` is set to that package when the source builds multiple packages).

          When `true`, this behaves similar to `dh_auto_install --destdir=debian/PACKAGE`.
        """
            ),
        ),
    ],
    DebputyParsedContentStandardConditional: [
        StandardParserAttributeDocumentation(
            frozenset(["when"]),
            textwrap.dedent(
                """\
            A condition as defined in [Conditional rules](${MANIFEST_FORMAT_DOC}#conditional-rules).

            The conditional will disable the entire rule when the conditional evaluates to false.
        """
            ),
            # Last
            sort_category=9999,
        ),
    ],
}


def docs_from(
    *ts: Any,
    exclude_attributes: Container[str] = frozenset(),
) -> Iterable[ParserAttributeDocumentation]:
    """Provide standard attribute documentation from existing types

    This is a work-around for `apply_standard_attribute_documentation` requiring python3.12.
    If you can assume python3.12, use `apply_standard_attribute_documentation` instead.
    """
    for t in ts:
        attrs = _STD_ATTR_DOCS.get(t)
        if attrs is None:
            raise ValueError(f"No standard documentation for {str(t)}")
        for attr in attrs:
            if any(a in exclude_attributes for a in attrs):
                continue
            yield attr



import stat
import os
from typing import Iterator

import pytest

from debputy.plugin.api.test_api import (
    initialize_plugin_under_test,
    build_virtual_file_system,
    package_metadata_context,
)

STUB_CMD = os.path.join(os.path.dirname(__file__), "perl-ssl_test.sh")


@pytest.fixture(scope="session")
def perl_ssl_stub_cmd() -> Iterator[None]:
    os.environ["_PERL_SSL_DEFAULTS_TEST_PATH"] = STUB_CMD
    mode = stat.S_IMODE(os.stat(STUB_CMD).st_mode)
    if (mode & 0o500) != 0o500:
        os.chmod(STUB_CMD, mode | 0o500)
    yield
    try:
        del os.environ["_PERL_SSL_DEFAULTS_TEST_PATH"]
    except KeyError:
        pass


def test_perl_openssl(perl_ssl_stub_cmd) -> None:
    plugin = initialize_plugin_under_test()
    fs = build_virtual_file_system([])
    context = package_metadata_context(package_fields={"Architecture": "all"})
    metadata = plugin.run_metadata_detector("perl-openssl-abi", fs, context)
    assert metadata.substvars["perl:Depends"] == "perl-openssl-abi-3"



import textwrap

import pytest

from debputy.lsp.lsp_debian_tests_control import _lint_debian_tests_control
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from lint_tests.lint_tutil import (
    requires_levenshtein,
    LintWrapper,
)
from debputy.lsprotocol.types import DiagnosticSeverity


@pytest.fixture
def line_linter(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
) -> LintWrapper:
    return LintWrapper(
        "/nowhere/debian/tests/control",
        _lint_debian_tests_control,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
    )


def test_dtctrl_lint_tests_vs_test_command(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Depends:
     @,
     python3-all,
     python3-pytest,
     python3-pytest-mock,
     python3-pytest-xvfb,
     xauth,
     xvfb,
    Restrictions: allow-stderr

    Tests: foo
    Test-Command: bar
    Depends:
     @,
     python3-all,
     python3-pytest,
     python3-pytest-mock,
     python3-pytest-xvfb,
     xauth,
     xvfb,
    Restrictions: allow-stderr
    """
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    print(diagnostics)
    assert len(diagnostics) == 2
    first_error, second_error = diagnostics

    msg = 'Stanza must have either a "Tests" or a "Test-Command" field'
    assert first_error.message == msg
    assert f"{first_error.range}" == "0:0-8:0"
    assert first_error.severity == DiagnosticSeverity.Error

    msg = 'Stanza cannot have both a "Tests" and a "Test-Command" field'
    assert second_error.message == msg
    assert f"{second_error.range}" == "10:0-11:0"
    assert second_error.severity == DiagnosticSeverity.Error


@requires_levenshtein
def test_dtctrl_lint_live_example_silx(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(
        """\
    Tests: no-opencl
    Depends:
     @,
     python3-all,
     python3-pytest,
     python3-pytest-mock,
     python3-pytest-xvfb,
     xauth,
     xvfb,
    Restrictions: allow-stderr

    Tests: opencl
    Depends:
     @,
     clinfo,
     python3-all,
     python3-pytest,
     python3-pytest-mock,
     python3-pytest-xvfb,
     xauth,
     xvfb,
    Architecture: !i386
    Restrictions: allow-stderr

    Test-Command: xvfb-run -s "-screen 0 1024x768x24 -ac +extension GLX +render -noreset" sh debian/tests/gui
    Depends:
     mesa-utils,
     silx,
     xauth,
     xvfb,
    Restrictions: allow-stderr
    """
    ).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    print(diagnostics)
    assert not diagnostics



from typing import TYPE_CHECKING

import pytest
from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.util import setup_logging

if TYPE_CHECKING:
    import lsprotocol.types as types

else:
    import debputy.lsprotocol.types as types

try:
    from pygls.server import LanguageServer

    HAS_PYGLS = True
except ImportError:
    HAS_PYGLS = False


@pytest.fixture(scope="session", autouse=True)
def enable_logging() -> None:
    setup_logging(log_only_to_stderr=True, reconfigure_logging=True)


@pytest.fixture
def lsp_dctrl_parser(
    dpkg_arch_query: DpkgArchTable,
    amd64_dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
    no_profiles_or_build_options: DebBuildOptionsAndProfiles,
) -> DctrlParser:
    return DctrlParser(
        frozenset(),
        frozenset(),
        True,
        True,
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        no_profiles_or_build_options,
    )


@pytest.fixture()
def ls(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lsp_dctrl_parser: DctrlParser,
) -> "DebputyLanguageServer":
    if not HAS_PYGLS:
        pytest.skip("Missing pygls")
    ls = DebputyLanguageServer("debputy", "v<test>")
    ls.lsp.lsp_initialize(
        types.InitializeParams(
            types.ClientCapabilities(
                general=types.GeneralClientCapabilities(
                    position_encodings=[types.PositionEncodingKind.Utf32],
                )
            )
        )
    )
    ls.plugin_feature_set = debputy_plugin_feature_set
    ls.dctrl_parser = lsp_dctrl_parser
    return ls



from typing import Container, Optional, Mapping

from debputy.plugin.api.spec import (
    DebputyIntegrationMode,
    INTEGRATION_MODE_DH_DEBPUTY_RRR,
    INTEGRATION_MODE_DH_DEBPUTY,
    INTEGRATION_MODE_FULL,
)


def determine_debputy_integration_mode(
    source_fields: Mapping[str, str],
    all_sequences: Container[str],
) -> Optional[DebputyIntegrationMode]:

    if source_fields.get("Build-Driver", "").lower() == "debputy":
        return INTEGRATION_MODE_FULL

    has_zz_debputy = "zz-debputy" in all_sequences or "debputy" in all_sequences
    has_zz_debputy_rrr = "zz-debputy-rrr" in all_sequences
    has_any_existing = has_zz_debputy or has_zz_debputy_rrr
    if has_zz_debputy_rrr:
        return INTEGRATION_MODE_DH_DEBPUTY_RRR
    if has_any_existing:
        return INTEGRATION_MODE_DH_DEBPUTY
    if source_fields.get("Source", "") == "debputy":
        # Self-hosting. We cannot set the Build-Driver field since that creates a self-circular dependency loop
        return INTEGRATION_MODE_FULL
    return None



import dataclasses
from typing import (
    Iterator,
    Union,
    Optional,
    List,
    Tuple,
    Mapping,
    get_origin,
    get_args,
    Any,
    Type,
    TypeVar,
    TYPE_CHECKING,
    Iterable,
    Container,
    Literal,
    FrozenSet,
    cast,
)

from debputy.yaml.compat import CommentedBase

from debputy.manifest_parser.exceptions import ManifestParseException

if TYPE_CHECKING:
    from debputy.manifest_parser.parser_data import ParserContextData
    from debputy.plugin.api.spec import DebputyIntegrationMode, PackageTypeSelector


MP = TypeVar("MP", bound="DebputyParseHint")
StrOrInt = Union[str, int]
AttributePathAliasMapping = Mapping[
    StrOrInt, Tuple[StrOrInt, Optional["AttributePathAliasMapping"]]
]
LineReportKind = Literal["key", "value", "container"]


_PACKAGE_TYPE_DEB_ONLY = frozenset(["deb"])
_ALL_PACKAGE_TYPES = frozenset(["deb", "udeb"])


def resolve_package_type_selectors(
    package_type: "PackageTypeSelector",
) -> FrozenSet[str]:
    if package_type is _ALL_PACKAGE_TYPES or package_type is _PACKAGE_TYPE_DEB_ONLY:
        return cast("FrozenSet[str]", package_type)
    if isinstance(package_type, str):
        return (
            _PACKAGE_TYPE_DEB_ONLY
            if package_type == "deb"
            else frozenset([package_type])
        )
    else:
        return frozenset(package_type)


class AttributePath:
    __slots__ = ("parent", "container", "name", "alias_mapping", "path_hint")

    def __init__(
        self,
        parent: Optional["AttributePath"],
        key: Optional[Union[str, int]],
        *,
        container: Optional[Any] = None,
        alias_mapping: Optional[AttributePathAliasMapping] = None,
    ) -> None:
        self.parent = parent
        self.container = container
        self.name = key
        self.path_hint: Optional[str] = None
        self.alias_mapping = alias_mapping

    @classmethod
    def root_path(cls, container: Optional[Any]) -> "AttributePath":
        return AttributePath(None, None, container=container)

    @classmethod
    def builtin_path(cls) -> "AttributePath":
        return AttributePath(None, "$builtin$")

    @classmethod
    def test_path(cls) -> "AttributePath":
        return AttributePath(None, "$test$")

    def __bool__(self) -> bool:
        return self.name is not None or self.parent is not None

    def copy_with_path_hint(self, path_hint: str) -> "AttributePath":
        p = self.__class__(self.parent, self.name, alias_mapping=self.alias_mapping)
        p.path_hint = path_hint
        return p

    def path_segments(self) -> Iterable[Union[str, int]]:
        segments = list(self._iter_path())
        segments.reverse()
        yield from (s.name for s in segments)

    def _resolve_path(self, report_kind: LineReportKind) -> str:
        parent = self.parent
        key = self.name
        if report_kind == "container":
            key = parent.name if parent else None
            parent = parent.parent if parent else None
        container = parent.container if parent is not None else None

        if isinstance(container, CommentedBase):
            lc = container.lc
            try:
                if isinstance(key, str):
                    if report_kind == "key":
                        lc_data = lc.key(key)
                    else:
                        lc_data = lc.value(key)
                else:
                    lc_data = lc.item(key)
            except (AttributeError, RuntimeError, LookupError, TypeError):
                lc_data = None
        else:
            lc_data = None

        segments = list(self._iter_path())
        segments.reverse()
        parts: List[str] = []
        path_hint = None

        for s in segments:
            k = s.name
            s_path_hint = s.path_hint
            if s_path_hint is not None:
                path_hint = s_path_hint
            if isinstance(k, int):
                parts.append(f"[{k}]")
            elif k is not None:
                if parts:
                    parts.append(".")
                parts.append(k)

        if lc_data is not None:
            line_pos, col = lc_data
            # Translate 0-based (index) to 1-based (line number)
            line_pos += 1
            parts.append(f" [Line {line_pos} column {col}]")

        elif path_hint:
            parts.append(f" <Search for: {path_hint}>")
        if not parts:
            return "document root"
        return "".join(parts)

    @property
    def path_container_lc(self) -> str:
        return self._resolve_path("container")

    @property
    def path_key_lc(self) -> str:
        return self._resolve_path("key")

    @property
    def path(self) -> str:
        return self._resolve_path("value")

    def __str__(self) -> str:
        return self.path

    def __getitem__(self, item: Union[str, int]) -> "AttributePath":
        alias_mapping = None
        if self.alias_mapping:
            match = self.alias_mapping.get(item)
            if match:
                item, alias_mapping = match
                if item == "":
                    # Support `sources[0]` mapping to `source` by `sources -> source` and `0 -> ""`.
                    return AttributePath(
                        self.parent,
                        self.name,
                        alias_mapping=alias_mapping,
                        container=self.container,
                    )
        container = self.container
        if container is not None:
            try:
                child_container = self.container[item]
            except (AttributeError, RuntimeError, LookupError, TypeError):
                child_container = None
        else:
            child_container = None
        return AttributePath(
            self,
            item,
            alias_mapping=alias_mapping,
            container=child_container,
        )

    def _iter_path(self) -> Iterator["AttributePath"]:
        current = self
        yield current
        while True:
            parent = current.parent
            if not parent:
                break
            current = parent
            yield current


def check_integration_mode(
    path: AttributePath,
    parser_context: Optional["ParserContextData"] = None,
    expected_debputy_integration_mode: Optional[
        Container["DebputyIntegrationMode"]
    ] = None,
) -> None:
    if expected_debputy_integration_mode is None:
        return
    if parser_context is None:
        raise AssertionError(
            f"Cannot use integration mode restriction when parsing {path.path} since it is not parsed in the manifest context"
        )
    if parser_context.debputy_integration_mode not in expected_debputy_integration_mode:
        raise ManifestParseException(
            f"The attribute {path.path} cannot be used as it is not allowed for"
            f" the current debputy integration mode ({parser_context.debputy_integration_mode})."
            f" Please remove the manifest definition or change the integration mode"
        )


@dataclasses.dataclass(slots=True, frozen=True)
class _SymbolicModeSegment:
    base_mode: int
    base_mask: int
    cap_x_mode: int
    cap_x_mask: int

    def apply(self, current_mode: int, is_dir: bool) -> int:
        if current_mode & 0o111 or is_dir:
            chosen_mode = self.cap_x_mode
            mode_mask = self.cap_x_mask
        else:
            chosen_mode = self.base_mode
            mode_mask = self.base_mask
        # set ("="): mode mask clears relevant segment and current_mode are the desired bits
        # add ("+"): mode mask keeps everything and current_mode are the desired bits
        # remove ("-"): mode mask clears relevant bits and current_mode are 0
        return (current_mode & mode_mask) | chosen_mode


def _symbolic_mode_bit_inverse(v: int) -> int:
    # The & part is necessary because otherwise python narrows the inversion to the minimum number of bits
    # required, which is not what we want.
    return ~v & 0o7777


def parse_symbolic_mode(
    symbolic_mode: str,
    attribute_path: Optional[AttributePath],
) -> Iterator[_SymbolicModeSegment]:
    sticky_bit = 0o01000
    setuid_bit = 0o04000
    setgid_bit = 0o02000
    mode_group_flag = 0o7
    subject_mask_and_shift = {
        "u": (mode_group_flag << 6, 6),
        "g": (mode_group_flag << 3, 3),
        "o": (mode_group_flag << 0, 0),
    }
    bits = {
        "r": (0o4, 0o4),
        "w": (0o2, 0o2),
        "x": (0o1, 0o1),
        "X": (0o0, 0o1),
        "s": (0o0, 0o0),  # Special-cased below (it depends on the subject)
        "t": (0o0, 0o0),  # Special-cased below
    }
    modifiers = {
        "+",
        "-",
        "=",
    }
    in_path = f" in {attribute_path.path}" if attribute_path is not None else ""
    for orig_part in symbolic_mode.split(","):
        base_mode = 0
        cap_x_mode = 0
        part = orig_part
        subjects = set()
        while part and part[0] in ("u", "g", "o", "a"):
            subject = part[0]
            if subject == "a":
                subjects = {"u", "g", "o"}
            else:
                subjects.add(subject)
            part = part[1:]
        if not subjects:
            subjects = {"u", "g", "o"}

        if part and part[0] in modifiers:
            modifier = part[0]
        elif not part:
            raise ValueError(
                f'Invalid symbolic mode{in_path}: expected [+-=] to be present (from "{orig_part}")'
            )
        else:
            raise ValueError(
                f'Invalid symbolic mode{in_path}: Expected "{part[0]}" to be one of [+-=]'
                f' (from "{orig_part}")'
            )
        part = part[1:]
        s_bit_seen = False
        t_bit_seen = False
        while part and part[0] in bits:
            if part == "s":
                s_bit_seen = True
            elif part == "t":
                t_bit_seen = True
            elif part in ("u", "g", "o"):
                raise NotImplementedError(
                    f"Cannot parse symbolic mode{in_path}: Sorry, we do not support referencing an"
                    " existing subject's permissions (a=u) in symbolic modes."
                )
            else:
                matched_bits = bits.get(part[0])
                if matched_bits is None:
                    valid_bits = "".join(bits)
                    raise ValueError(
                        f'Invalid symbolic mode{in_path}: Expected "{part[0]}" to be one of the letters'
                        f' in "{valid_bits}" (from "{orig_part}")'
                    )
                base_mode_bits, cap_x_mode_bits = bits[part[0]]
                base_mode |= base_mode_bits
                cap_x_mode |= cap_x_mode_bits
            part = part[1:]

        if part:
            raise ValueError(
                f'Invalid symbolic mode{in_path}: Could not parse "{part[0]}" from "{orig_part}"'
            )

        final_base_mode = 0
        final_cap_x_mode = 0
        segment_mask = 0
        for subject in subjects:
            mask, shift = subject_mask_and_shift[subject]
            segment_mask |= mask
            final_base_mode |= base_mode << shift
            final_cap_x_mode |= cap_x_mode << shift
        if modifier == "=":
            segment_mask |= setuid_bit if "u" in subjects else 0
            segment_mask |= setgid_bit if "g" in subjects else 0
            segment_mask |= sticky_bit if "o" in subjects else 0
        if s_bit_seen:
            if "u" in subjects:
                final_base_mode |= setuid_bit
                final_cap_x_mode |= setuid_bit
            if "g" in subjects:
                final_base_mode |= setgid_bit
                final_cap_x_mode |= setgid_bit
        if t_bit_seen:
            final_base_mode |= sticky_bit
            final_cap_x_mode |= sticky_bit
        if modifier == "+":
            final_base_mask = ~0
            final_cap_x_mask = ~0
        elif modifier == "-":
            final_base_mask = _symbolic_mode_bit_inverse(final_base_mode)
            final_cap_x_mask = _symbolic_mode_bit_inverse(final_cap_x_mode)
            final_base_mode = 0
            final_cap_x_mode = 0
        elif modifier == "=":
            # FIXME: Handle "unmentioned directory's setgid/setuid bits"
            inverted_mask = _symbolic_mode_bit_inverse(segment_mask)
            final_base_mask = inverted_mask
            final_cap_x_mask = inverted_mask
        else:
            raise AssertionError(
                f"Unknown modifier in symbolic mode: {modifier} - should not have happened"
            )
        yield _SymbolicModeSegment(
            base_mode=final_base_mode,
            base_mask=final_base_mask,
            cap_x_mode=final_cap_x_mode,
            cap_x_mask=final_cap_x_mask,
        )


def unpack_type(
    orig_type: Any,
    parsing_typed_dict_attribute: bool,
) -> Tuple[Any, Optional[Any], Tuple[Any, ...]]:
    raw_type = orig_type
    origin = get_origin(raw_type)
    args = get_args(raw_type)
    if not parsing_typed_dict_attribute and repr(origin) in (
        "typing.NotRequired",
        "typing.Required",
    ):
        raise ValueError(
            f"The Required/NotRequired attributes cannot be used outside typed dicts,"
            f" the type that triggered the error: {orig_type}"
        )

    while repr(origin) in ("typing.NotRequired", "typing.Required"):
        if len(args) != 1:
            raise ValueError(
                f"The type {raw_type} should have exactly one type parameter"
            )
        raw_type = args[0]
        origin = get_origin(raw_type)
        args = get_args(raw_type)

    assert not isinstance(raw_type, tuple)

    return raw_type, origin, args


def find_annotation(
    annotations: Tuple[Any, ...],
    anno_class: Type[MP],
) -> Optional[MP]:
    m = None
    for anno in annotations:
        if isinstance(anno, anno_class):
            if m is not None:
                raise ValueError(
                    f"The annotation {anno_class.__name__} was used more than once"
                )
            m = anno
    return m



import dataclasses
from typing import (
    TypeVar,
    TYPE_CHECKING,
    Sequence,
    Tuple,
    Mapping,
    Dict,
    Optional,
    TypedDict,
    NotRequired,
    List,
    MutableMapping,
)

if TYPE_CHECKING:
    from debputy.plugin.api import VirtualPath
    from debputy.filesystem_scan import FSPath

    VP = TypeVar("VP", VirtualPath, FSPath)
    S = TypeVar("S", str, bytes)
else:
    VP = TypeVar("VP", "VirtualPath", "FSPath")
    S = TypeVar("S", str, bytes)


class EnvironmentModificationSerialized(TypedDict):
    replacements: NotRequired[Dict[str, str]]
    removals: NotRequired[List[str]]


@dataclasses.dataclass(slots=True, frozen=True)
class EnvironmentModification:
    replacements: Sequence[Tuple[str, str]] = tuple()
    removals: Sequence[str] = tuple()

    def __bool__(self) -> bool:
        return bool(self.removals) or bool(self.replacements)

    def combine(
        self,
        other: "Optional[EnvironmentModification]",
    ) -> "EnvironmentModification":
        if not other:
            return self
        existing_replacements = {k: v for k, v in self.replacements}
        extra_replacements = {
            k: v
            for k, v in other.replacements
            if k not in existing_replacements or existing_replacements[k] != v
        }
        seen_removals = set(self.removals)
        extra_removals = [r for r in other.removals if r not in seen_removals]

        if not extra_replacements and isinstance(self.replacements, tuple):
            new_replacements = self.replacements
        else:
            combined_replacements = []
            for k, v in existing_replacements.items():
                if k not in extra_replacements:
                    combined_replacements.append((k, v))

            for k, v in other.replacements:
                if k in extra_replacements:
                    combined_replacements.append((k, v))

            new_replacements = tuple(combined_replacements)

        if not extra_removals and isinstance(self.removals, tuple):
            new_removals = self.removals
        else:
            combined_removals = list(self.removals)
            combined_removals.extend(extra_removals)
            new_removals = tuple(combined_removals)

        if self.replacements is new_replacements and self.removals is new_removals:
            return self

        return EnvironmentModification(
            new_replacements,
            new_removals,
        )

    def update_inplace(self, env: MutableMapping[str, str]) -> None:
        for k, v in self.replacements:
            existing_value = env.get(k)
            if v == existing_value:
                continue
            env[k] = v

        for k in self.removals:
            if k not in env:
                continue
            del env[k]

    def compute_env(self, base_env: Mapping[str, str]) -> Mapping[str, str]:
        updated_env: Optional[Dict[str, str]] = None
        for k, v in self.replacements:
            existing_value = base_env.get(k)
            if v == existing_value:
                continue

            if updated_env is None:
                updated_env = dict(base_env)
            updated_env[k] = v

        for k in self.removals:
            if k not in base_env:
                continue
            if updated_env is None:
                updated_env = dict(base_env)
            del updated_env[k]

        if updated_env is not None:
            return updated_env
        return base_env
